"""A module containing the helical torsion spring class"""
from math import pi, sqrt

from me_toolbox.fatigue import FailureCriteria
from me_toolbox.springs import Spring
from me_toolbox.tools import percent_to_decimal


class HelicalTorsionSpring(Spring):
    """A Helical torsion spring object"""

    def __repr__(self):
        return f"HelicalTorsionSpring(max_moment={self.max_moment}, " \
               f"wire_diameter{self.wire_diameter}, spring_diameter={self.diameter}, " \
               f"leg1={self.leg1}, leg2={self.leg2}, " \
               f"ultimate_tensile_strength={self.ultimate_tensile_strength}, " \
               f"yield_percent={self.yield_percent}, shear_modulus={self.shear_modulus}, " \
               f"elastic_modulus={self.elastic_modulus}, spring_rate={self.spring_rate}, " \
               f"radius={self.radius}, arbor_diameter={self.arbor_diameter}, " \
               f"shot_peened={self.shot_peened}, density={self.density}, " \
               f"working_frequency={self.working_frequency})"

    def __str__(self):
        return f"HelicalTorsionSpring(d={self.wire_diameter}, D={self.diameter}, " \
               f"k={self.spring_rate})"

    def __init__(self, max_moment, wire_diameter, spring_diameter, leg1, leg2,
                 ultimate_tensile_strength, yield_percent, shear_modulus, elastic_modulus,
                 spring_rate, radius=None, arbor_diameter=None, shot_peened=False, density=None,
                 working_frequency=None):
        """Instantiate helical torsion spring object with the given parameters

        :param float  max_moment: The maximum load on the spring
        :param float  wire_diameter: spring wire diameter
        :param float  spring_diameter: spring diameter measured from
            the center point of the wire diameter
        :param float ultimate_tensile_strength: Ultimate tensile strength of the material
        :param float leg1: spring leg
        :param float leg2: spring leg
        :param float shear_modulus: Spring's material shear modulus
        :param float elastic_modulus: Spring's material elastic modulus
        :param float yield_percent: Used to estimate the spring's yield stress
        :param float spring_rate: K - spring rate
        :param float arbor_diameter: the diameter of the pin going through the spring
        :param float radius: The distance of applied force from the center
        :param bool shot_peened: if True adds to fatigue strength
        :param float or None density: Spring's material density
            (used for buckling and weight calculations)
        :param float or None working_frequency: the spring working frequency
            (used for fatigue calculations)

        :returns: HelicalTorsionSpring
        """
        max_force = max_moment / radius if radius is not None else None

        super().__init__(max_force, wire_diameter, spring_diameter, spring_rate,
                         ultimate_tensile_strength, shear_modulus, elastic_modulus,
                         shot_peened, density, working_frequency)

        self.radius = radius  # Needs a better name
        self.max_moment = max_moment
        self.yield_percent = yield_percent
        self.leg1 = leg1
        self.leg2 = leg2
        self.arbor_diameter = arbor_diameter

    def check_design(self):
        raise NotImplementedError("check_design is not implemented yet for HelicalTorsionSpring")

    @property
    def diameter_after_deflection(self):
        """ After deflection The spring diameter is shrinking it is important to know the new
        diameter to check if it conflicts with the arbor size

        :return: The spring diameter after deflection
        :rtype: float
        """
        Nb = self.body_coils
        return (Nb * self.diameter) / (Nb + self.calc_angular_deflection(self.max_moment, False))

    @property
    def clearance(self):
        """Diametrical Clearance between the spring after deflection and the arbor"""
        if self.arbor_diameter is not None:
            ID = self.diameter_after_deflection - self.wire_diameter
            return ID - self.arbor_diameter
        else:
            return "The pin diameter was not given"

    @property
    def length(self):
        """ The length of the spring - L """
        return self.wire_diameter * self.body_coils

    @property
    def active_coils(self):
        """Number of active coils - Na

        Note: according to the analytical calculation the 67.8584 should be 64 (from the second
        moment of inertia of a round wire) but 67.8584 fits better to the experimental results,
        probably due to friction between the coils and arbor.

        :returns: number of active coils
        :rtype: float
        """
        D = self.diameter
        d = self.wire_diameter
        active_coils = (d ** 4 * self.elastic_modulus) / (67.8584 * D * self.spring_rate)
        return active_coils

    @property
    def body_coils(self):
        """Total number of coils

        :returns: number of body coils
        :rtype: float
        """
        return self.active_coils - ((self.leg1 + self.leg2) / (3 * pi * self.diameter))

    @property
    def partial_turn(self):
        """Partial number of coils (β/360)

        :returns: Partial number of coils
        :rtype: float
        """
        return self.body_coils - int(self.body_coils)

    @property
    def yield_strength(self):
        """ Sy - yield strength
        (shear_yield_stress = % * ultimate_tensile_strength))
        """
        try:
            return percent_to_decimal(self.yield_percent) * self.ultimate_tensile_strength
        except TypeError:
            return self.yield_percent * self.ultimate_tensile_strength

    @property
    def factor_Ki(self):
        """Inner fibers stress correction factor

        :returns:stress concentration factor ki
        :rtype: float
        """
        index = self.spring_index
        return (4 * index ** 2 - index - 1) / (4 * index * (index - 1))

    @property
    def factor_Ko(self):
        """Outer fiber stress correction factor. in light that factor_Ko is always less than
        factor_ki we don't use it in the stress estimation, but it is brought here
        for the sake of completion

        :returns:stress concentration factor Ko
        :rtype: float
        """
        index = self.spring_index
        return (4 * index ** 2 - index - 1) / (4 * index * (index - 1))

    @property
    def max_stress(self):
        """The normal stress due to bending and axial loads

        :returns: Normal stress
        :rtype: float 
        """
        return self.calc_max_stress(self.max_moment)

    def calc_max_stress(self, moment):
        """Calculates the normal stress based on the moment given
        NOTE: The calculation is for round wire torsion spring.

        :param float moment: Working force of the spring

        :returns: normal stress
        :rtype: float 
        """
        return self.factor_Ki * ((32 * moment) / (pi * self.wire_diameter ** 3))

    @property
    def natural_frequency(self):
        raise NotImplementedError("natural_frequency has not been implemented yet "
                                  "for HelicalTorsionSpring")

    @property
    def max_angular_deflection(self):
        """The total angular deflection due to the max moment
        this deflection consists of the angular deflection
        of the coil body and from the end deflection of cantilever
        for *each* leg.

        :returns: Max angular deflection
        :rtype: float 
        """
        return self.calc_angular_deflection(self.max_moment)

    def calc_angular_deflection(self, moment, total_deflection=True):
        """Calculates the total angular deflection based on the moment given,
        If the total flag is True than the total angular deflection is calculated,
        if False only the deflection of the coil body is calculated (without the legs)

        Note: according to the analytical calculation the 67.8584 should be 64 (from the second
        moment of inertia of a round wire) but 67.8584 fits better to the experimental results,
        probably due to friction between the coils and arbor.

        :param float  moment: Working moment of the spring
        :param bool total_deflection: total or partial deflection

        :returns: Total angular deflection in radians
        :rtype: float 
        """
        d = self.wire_diameter
        D = self.diameter
        E = self.elastic_modulus
        l1 = self.leg1
        l2 = self.leg2
        Nb = self.body_coils
        legs_deflection_part = (l1 + l2)/(3*pi*D) if total_deflection else 0
        return ((67.8584 * moment * D) / (d ** 4 * E)) * (Nb + legs_deflection_part)

    @property
    def weight(self):
        """Return's the spring *active coils* weight according to the specified density

        :returns: Spring weight
        :type: float
        """
        area = 0.25 * pi * (self.wire_diameter * 1e-3) ** 2  # cross-section area
        length = pi * self.diameter * 1e-3  # the circumference of the spring
        coil_volume = area * length
        if self.density is not None:
            return (coil_volume * self.body_coils + (self.leg1 + self.leg2) * area) * self.density
        else:
            raise ValueError(f"Can't calculate weight, no density is specified")

    @property
    def static_safety_factor(self):  # pylint: disable=unused-argument
        """ Returns the static safety factor

        :returns: Spring's safety factor
        :type: float
        """
        return self.yield_strength / self.max_stress

    def fatigue_analysis(self, max_moment, min_moment, reliability,
                         criterion='gerber', verbose=False, metric=True):
        """ Returns safety factors for fatigue and
        for first cycle according to Langer failure criteria.

        :param float max_moment: Maximal max_force acting on the spring
        :param float min_moment: Minimal max_force acting on the spring
        :param float reliability: in percentage
        :param str criterion: fatigue criterion
        :param bool verbose: print more details
        :param bool metric: Metric or imperial

        :returns: static and dynamic safety factor
        :rtype: tuple[float, float]
        """
        # calculating mean and alternating forces
        alt_moment = abs(max_moment - min_moment) / 2
        mean_moment = (max_moment + min_moment) / 2

        # calculating mean and alternating stresses
        alt_stress = self.calc_max_stress(alt_moment)
        mean_stress = self.calc_max_stress(mean_moment)

        Sse = self.shear_endurance_limit(reliability, metric)
        Se = Sse / 0.577  # based on the distortion energy method
        Sut = self.ultimate_tensile_strength
        Sy = self.yield_strength
        nf, nl = FailureCriteria.get_safety_factors(Sy, Sut, Se, alt_stress, mean_stress, criterion)
        if verbose:
            print(f"Alternating moment = {alt_moment}, Mean moment = {mean_moment}\n"
                  f"Alternating stress = {alt_stress}, Mean stress = {mean_stress}\n"
                  f"Sse = {Sse}, Se= {Se}")
        return nf, nl

    def min_wire_diameter(self, safety_factor, Ap, m, spring_diameter=None, spring_index=None):
        """The minimal wire diameter for given safety factor
        in order to avoid failure, according to the spring parameters

        Note: In order for the calculation to succeed the
            spring diameter or the spring index should be known.

        :param float safety_factor: static safety factor
        :param float spring_diameter: The spring diameter
        :param float spring_index: The spring index

        :returns: The minimal wire diameter
        :rtype: float
        """
        factor_k, temp_k = 1.1, 0
        diam = 0
        while abs(factor_k - temp_k) > 1e-4:
            # waiting for k to converge
            diam = ((32 * self.max_moment * factor_k * safety_factor) / (
                    self.yield_percent * Ap * pi)) ** (
                           1 / (3 - m))
            temp_k = factor_k
            if spring_diameter is not None:
                D = spring_diameter
                factor_k = (4 * D ** 2 - D * diam - diam ** 2) / (4 * D * (D - diam))
            elif spring_index is not None:
                c = spring_index
                factor_k = (4 * c ** 2 - c - 1) / (4 * c * (c - 1))
            else:
                print("Need to know spring index or spring diameter to calculate wire diameter")
        return diam

    def min_spring_diameter(self, safety_factor, wire_diameter):
        """return the minimum spring diameter to avoid static failure
        according to the specified safety factor.

        :param float safety_factor: static safety factor
        :param float wire_diameter: Spring's wire diameter

        :returns: The minimal spring diameter
        :rtype: float
        """
        d = wire_diameter
        Sy = self.yield_strength
        M = self.max_moment
        alpha = 4 * (Sy * pi * d ** 3 - 32 * M * safety_factor)
        beta = -d * (4 * Sy * pi * d ** 3 + 32 * M * safety_factor)
        gamma = 32 * M * safety_factor * d ** 2
        return (-beta + sqrt(beta ** 2 - 4 * alpha * gamma)) / (2 * alpha)

    @staticmethod
    def calc_spring_rate(wire_diameter, spring_diameter, active_coils, elastic_modulus):
        """Estimate spring constant from geometric and material properties

        :returns: The spring constant [Nm/rad]
        :rtype: float
        """
        d = wire_diameter
        D = spring_diameter
        return (d ** 4 * elastic_modulus) / (67.8584 * D * active_coils)



"""Module containing the Gear class"""

from math import log, sqrt, pi, tan, radians
import os
import numpy as np

from me_toolbox.tools import table_interpolation


class Gear:
    """a Gear object"""
    def __repr__(self):
        return f"{self.__class__.__name__}(m={self.modulus}, N={self.teeth_num}, " \
               f"\N{GREEK SMALL LETTER PHI}={self.pressure_angle}, b={self.width})"

    def __init__(self, modulus, teeth_num, rpm, Qv, width, bearing_span, pinion_offset,
                 enclosure, hardness, pressure_angle, grade, work_hours, number_of_cycles,
                 crowned, adjusted, sensitive_use, nitriding, case_carb, material):

        self.modulus = modulus
        self.teeth_num = teeth_num
        self.rpm = rpm
        self.Qv = Qv
        self.width = width
        self.bearing_span = bearing_span
        self.pinion_offset = pinion_offset
        self.enclosure = enclosure
        self.hardness = hardness
        self.pressure_angle = pressure_angle
        self.grade = grade
        self.crowned = crowned
        self.adjusted = adjusted
        self.sensitive_use = sensitive_use
        self.nitriding = nitriding
        self.case_carb = case_carb
        self.material = material

        if work_hours != 0:
            self.work_hours = work_hours
        if number_of_cycles != 0:
            self.number_of_cycles = number_of_cycles
        if work_hours == 0 and number_of_cycles == 0:
            raise ValueError('work_hours or number_of_cycles not specified')
        self.contact_ratio = None
        self.maximum_velocity = None

        self.Zw = None

    @property
    def pitch_diameter(self):
        """Calculate pitch diameter

        :returns: Gear's pitch diameter
        :rtype: float
        """
        return self.teeth_num * self.modulus  # pitch diameter [mm]

    @property
    def KB(self):
        """Rim thickness factor, factor_KB is dependent on the number of teeth


        :returns: Gear's Rim thickness factor
        :rtype: float
        """
        mB = (0.5 * self.teeth_num - 1.25) / 2.25
        if mB < 1.2:
            K_B = 1.6 * log(2.242 / mB)
        else:
            K_B = 1
        return K_B

    @property
    def Kv(self):
        """Dynamic factor, Kv is dependent on the pitch diameter in [mm],
        the angular velocity in [rpm] and Qv (transmission accuracy grade number)

        note: the equations here are from "Shigley's Mechanical Engineering Design"
               in the AGMA standard the transmission accuracy grade number is Av instead of Qv
               and its value is inverted, high Qv is more accurate and high Av is less accurate

        :returns: Gear's Dynamic factor
        :rtype: float
        """
        B = 0.25 * (12 - self.Qv) ** (2 / 3)
        A = 50 + 56 * (1 - B)

        # maximum velocity
        self.maximum_velocity = ((A + (self.Qv - 3)) ** 2) / 200

        if 6 <= self.Qv <= 11:
            K_v = ((A + sqrt(200 * self.tangent_velocity)) / A) ** B
        elif self.Qv == 5:
            K_v = (50 + sqrt(200 * self.tangent_velocity)) / 50
        elif self.Qv == 12:
            K_v = 1
            self.maximum_velocity = "Qv=12 no maximum velocity"
        else:
            raise ValueError(f"at Kv factor: Qv={self.Qv} not in range (5<=Qv<=12)\n")
        return K_v

    @property
    def Ks(self):
        """Size factor, factor_Ks is dependent on the circular
        pitch (p=πm) which in turn depends on the modulus

        :returns: Gear's size factor
        :rtype: float
        """
        if self.pitch > 8:
            K_s = (1 / 1.189) * (self.pitch ** 0.097)
        else:
            K_s = 1
        return K_s

    @property
    def KH(self):
        """Load distribution factor, KH is dependent on: the shape of teeth (crowned),
        if teeth are adjusted after assembly (adjusted), the gear width in [mm],
        pitch diameter in [mm], bearing span (the distance between the bearings center lines)
        pinion offset (the distance from the bearing span center to the pinion mid-face)
        enclosure type (open gearing, commercial enclosed, precision enclosed,
        extra precision enclosed)

        :returns: Gear's load distribution factor
        :rtype: float
        """
        # K_Hmc - lead correction factor
        if self.crowned:
            K_Hmc = 0.8
        else:
            K_Hmc = 1

        # K_He - mesh alignment correction factor
        if self.adjusted:
            K_He = 0.8
        else:
            K_He = 1

        # K_Hpf - mesh alignment correction factor
        # gear width to diameter ratio
        ratio = self.width / (10 * self.pitch_diameter)
        if ratio < 0.05:
            ratio = 0.05

        # print("the width is", b)
        if self.width <= 25:
            K_Hpf = ratio - 0.025
        elif 25 < self.width <= 432:
            K_Hpf = ratio - 0.0375 + 0.000492 * self.width
        elif 432 < self.width <= 1020:
            K_Hpf = ratio - 0.1109 + 0.00815 * self.width - 0.000000353 * self.width ** 2
        else:
            raise ValueError(f"at KH factor: width={self.width} not in range, (width < 1020)\n")

        # K_Hpm - pinion proportion modifier
        s = self.bearing_span
        s1 = self.pinion_offset
        if (s1 / s) < 0.175:
            K_Hpm = 1
        else:
            K_Hpm = 1.1

        # K_Hma - mash alignment factor
        enclosure_type = {'open gearing': [2.47e-1, 0.657e-3, -1.186e-7],
                          'commercial enclosed': [1.27e-1, 0.622e-3, -1.69e-7],
                          'precision enclosed': [0.675e-1, 0.504e-3, -1.44e-7],
                          'extra precision enclosed': [0.380e-1, 0.402e-3, -1.27e-7]}
        K_Hma = (enclosure_type[self.enclosure][0] + self.width * enclosure_type[self.enclosure][1]
                 + enclosure_type[self.enclosure][2] * self.width ** 2)

        K_H = 1.0 + K_Hmc * (K_Hpf * K_Hpm + K_Hma * K_He)
        return K_H

    @property
    def St(self):
        """Bending safety factor, St is dependent on the gear's
         hardness in [HBN] and the material grade
         (material grade of the gear 1 for regular use 2 for military and sensitive uses)

        :returns: Gear's bending safety factor
        :rtype: float
        """
        if self.grade == 1:
            S_t = 0.533 * self.hardness + 88.3
        elif self.grade == 2:
            S_t = 0.703 * self.hardness + 113
        else:
            raise ValueError(f"at St factor: grade={self.grade} but only 1 or 2 are valid")

        return S_t

    @property
    def Sc(self):
        """Contact safety factor, Sc is dependent on the gear
        hardness in [HBN] and on the material grade
        (material grade of the gear 1 for regular use 2 for military and sensitive uses)

        :returns: Gear's contact safety factor
        :rtype: float
        """
        if self.grade == 1:
            S_c = 200 + 2.22 * self.hardness
        elif self.grade == 2:
            S_c = 237 + 2.41 * self.hardness
        else:
            raise ValueError(f"at Sc factor: grade={self.grade} but only 1 or 2 are valid")

        return S_c

    @property
    def ZR(self):
        """For now ZR is 1 according to the AGMA standard
        :rtype: int
        """
        return 1

    @property
    def YN(self):
        """Bending strength stress cycle factor

        :returns: Gear's bending strength stress cycle factor
        :rtype: float
        """
        # get the number of cycles of the gear
        N = self.cycles_or_hours()

        if N is None:
            return None

        # 1e2 <= N < 2e6
        low_cycle = {160: 2.3194 * N ** (-0.0538),
                     'Nitrided': 3.517 * N ** (-0.0817),
                     250: 4.9404 * N ** (-0.1045),
                     'Case carb': 6.1514 * N ** (-0.1192),
                     400: 9.4518 * N ** (-0.148)}

        # N>=2e6
        high_cycle = {True: 1.6831 * N ** (-0.0323),
                      False: 1.3558 * N ** (-0.0178)}

        try:
            if 1e2 <= N < 2e6 and self.nitriding:
                curve = 'Nitrided'
            elif 1e2 <= N < 2e6 and self.case_carb:
                curve = 'Case carb'
            else:
                curve = self.hardness

            if 1e2 <= N < 2e6:
                Y_N = low_cycle[curve]
            elif N >= 2e6:
                Y_N = high_cycle[self.sensitive_use]
            else:
                raise ValueError(f" at YN: the number of cycles is {N} "
                                 f"but the minimum number is {1e2} ")
            return Y_N
        except KeyError as bad_key:
            print(f"at YN: not valid hardness {bad_key}")
            return "Error"

    @property
    def ZN(self):
        """Calculating contact strength stress cycle factor

        :returns: Gear's contact strength stress cycle
        :rtype: float
        """
        # get the number of cycles of the gear
        N = self.cycles_or_hours()
        if N is None:
            return None

        # N < 3e6
        low_cycle = {True: 1.249 * N ** (-0.0138),  # nitrided
                     False: 2.466 * N ** (-0.056)}
        # N >= 3e6
        high_cycle = {True: 2.466 * N ** (-0.056),  # for sensitive use
                      False: 1.4488 * N ** (-0.023)}

        return low_cycle[self.nitriding] if N < 3e6 else high_cycle[self.sensitive_use]

    @staticmethod
    def Y_j(gear1, gear2):
        """Return Geometry factors for spur gear and pinion

        :returns: Gear's geometry factor
        :rtype: float
        """
        N1 = gear1.teeth_num
        N2 = gear2.teeth_num
        pressure_angle = gear1.pressure_angle

        # load table according to pressure angle
        if pressure_angle == 20:
            path = os.path.dirname(__file__) + "\\tables\\20deg - spur gear geometry factors.csv"
        elif pressure_angle == 25:
            path = os.path.dirname(__file__) + "\\tables\\25deg - spur gear geometry factors.csv"
        else:
            raise ValueError("at spur gear Yj Factor: pressure angle is wrong")

        data = np.genfromtxt(path, delimiter=',')
        # try:
        gear1.Yj = table_interpolation(N1, N2, data)
        gear2.Yj = table_interpolation(N2, N1, data)
        # except NotInRangeError as not_in_range:
        # print(f"Error: Teeth number of one of the gears ({not_in_range.num})
        # not in range {not_in_range.range_}")

    @property
    def tangent_velocity(self):
        """Convert tangent velocity to [m/s] from [rpm]

        :returns: Gear's tangent velocity
        :rtype: float
        """
        return (pi * self.pitch_diameter * self.rpm) / 60e3

    @property
    def pitch(self):
        """Calculate circular pitch

        :returns: Gear's circular pitch
        :rtype: float
        """
        return self.modulus * pi

    def cycles_or_hours(self):
        """Check if the gear number of cycles or work hours
        were input and return the number of cycle

        :returns: Gear's Dynamic factor
        :rtype: float or None

        :raise ValueError: number of cycles and work hours entered but they're not matching
            or no number of cycles or work hours entered
        """

        if self.contact_ratio is None:
            return None

        # if attribute don't exist assign zero
        cycles_number = self.__dict__.get("number_of_cycles", 0)
        working_hours = self.__dict__.get("work_hours", 0)

        if cycles_number == 0 and working_hours != 0:
            # calculate number of cycles
            return 60 * working_hours * self.rpm * self.contact_ratio
        elif cycles_number != 0 and working_hours == 0:
            # number of cycle is an input
            return cycles_number
        elif cycles_number != 0 and working_hours != 0:
            # if both were an input check if the match
            if cycles_number == 60 * working_hours * self.rpm * self.contact_ratio:
                return cycles_number
            else:
                raise ValueError("at YN factor: number of cycles and work"
                                 "hours entered but they're not matching")
        else:
            raise ValueError("at YN factor: no number of cycles or work hours entered")

    @staticmethod
    def calc_forces(gear, power):
        """Calculate forces on the gear

        :param SpurGear gear: gear object
        :param float power: power

        :returns: Wt - tangent max_force in [N], Wr - radial max_force in [N]
        :rtype: tuple[float, float]
        """

        Wt = (60e3 / pi) * (power / (gear.pitch_diameter * gear.rpm))
        Wr = Wt * tan(radians(gear.pressure_angle))
        return Wt, Wr

    @staticmethod
    def create_new_gear(gear2_prop):
        pass

    @staticmethod
    def format_properties(properties):
        pass

    def check_compatibility(self, gear):
        pass

    @staticmethod
    def ZI(gear1, gear2):
        pass

    def calc_centers_distance(self, gear_ratio):
        pass

    def optimization(self, transmission, optimize_feature='all', verbose=False):
        pass



from me_toolbox.fatigue.failure_criteria import FailureCriteria
from me_toolbox.fatigue.fatigue_analysis import FatigueAnalysis
from me_toolbox.fatigue.endurance_limit import EnduranceLimit
from me_toolbox.fatigue.stress import *




"""module containing the FatigueCriteria class
containing the Failure criteria as described in
Shigley's Mechanical Engineering design
"""
from sympy import sqrt


class FailureCriteria:
    """Bundling the fatigue criteria together"""

    @staticmethod
    def modified_goodman(ultimate_strength, endurance_limit, alt_eq_stress,
                         mean_eq_stress):
        """Safety factor according to modified Goodman failure criterion
        (a very common criterion)

        :returns: Safety factor
        :rtype: any

        :raises Exception: if mean equivalent stress is negative
        """
        if mean_eq_stress < 0:
            raise ValueError("Not valid when the mean equivalent stress is negative")

        return 1 / ((alt_eq_stress / endurance_limit) + (mean_eq_stress / ultimate_strength))

    @staticmethod
    def soderberg(yield_strength, endurance_limit, alt_eq_stress, mean_eq_stress):
        """Safety factor according to Soderberg failure criterion
        Soderberg criterion guards against any yielding, but is biased low (the safest criterion)

        :returns: Safety factor
        :rtype: any

        :raises Exception: if mean equivalent stress is negative
        """
        if mean_eq_stress < 0:
            raise ValueError("Not valid when the mean equivalent stress is negative")

        return 1 / ((alt_eq_stress / endurance_limit) + (mean_eq_stress / yield_strength))

    @staticmethod
    def gerber(ultimate_strength, endurance_limit, alt_eq_stress, mean_eq_stress):
        """Safety factor according to Gerber failure criterion
        (the most lenient criterion)

        :returns: Safety factor
        :rtype: any
        :raises Exception: if mean equivalent stress is negative
        """
        if mean_eq_stress < 0:
            raise ValueError("Not valid when the mean equivalent stress is negative")

        alpha = ultimate_strength / mean_eq_stress
        beta = alt_eq_stress / endurance_limit
        return 0.5 * alpha ** 2 * beta * (-1 + sqrt(1 + 4 * alpha ** (-2) * beta ** (-2)))

    @staticmethod
    def asme_elliptic(yield_strength, endurance_limit, alt_eq_stress,
                      mean_eq_stress):
        """Safety factor according to ASME Failure criterion

        :returns: Safety factor
        :rtype: any

        :raises Exception: if mean equivalent stress is negative
        """
        if mean_eq_stress < 0:
            raise ValueError("Not valid when the mean equivalent stress is negative")

        return sqrt(1 / ((alt_eq_stress / endurance_limit) ** 2 + (
                mean_eq_stress / yield_strength) ** 2))

    @staticmethod
    def langer_static_yield(yield_strength, alt_eq_stress, mean_eq_stress):
        """ Static safety factor according to Langer Failure criterion
        it's customary to use Langer, as an assessment to yielding in the first cycle

        :returns: Safety factor
        :rtype: any
        """
        if mean_eq_stress > 0:
            # stress is in the first quadrant of the alternating-mean stress plan
            return yield_strength / (alt_eq_stress + mean_eq_stress)
        else:
            # stress is in the second quadrant of the alternating-mean stress plan
            return yield_strength / (alt_eq_stress - mean_eq_stress)

    @staticmethod
    def get_safety_factors(yield_strength, ultimate_strength, endurance_limit,
                           alt_eq_stress, mean_eq_stress, criterion, verbose=False):
        """Returns dynamic and static safety factors
        according to the quadrant in the alternating-mean
        stress plain where the stresses are in

        Note: Takes into account the stress direction and offer alternative calculation
        in case the mean equivalent stress is negative
        :param str criterion: The criterion to use
        :param float yield_strength: The yield strength (Sy or Ssy)
        :param float ultimate_strength: The yield strength (Sut or Ssu)
        :param float endurance_limit: Modified endurance limit (Se)
        :param float alt_eq_stress: alternating stresses
        :param float mean_eq_stress: mean stresses
        :param bool verbose: Print the result

        :returns: dynamic and static safety factors
        :rtype: tuple[float, float]
        """

        if mean_eq_stress > 0:
            # stress is in the first quadrant of the alternating-mean stress plan
            args = (endurance_limit, alt_eq_stress, mean_eq_stress)
            safety_factors = {
                'modified goodman': FailureCriteria.modified_goodman(ultimate_strength, *args),
                'soderberg': FailureCriteria.soderberg(yield_strength, *args),
                'gerber': FailureCriteria.gerber(ultimate_strength, *args),
                'asme-elliptic': FailureCriteria.asme_elliptic(yield_strength, *args)}
            try:
                fatigue_safety_factor = safety_factors[criterion.lower()]

            except KeyError:
                raise Exception(f"Unknown criterion - {criterion}\n"
                                f"Available criteria are: 'Modified Goodman', 'Soderberg',"
                                f"'Gerber', 'ASME-elliptic'")

        else:
            # stress is in the second quadrant of the alternating-mean stress plan
            if verbose:
                print(f"NOTE: The mean stress = {mean_eq_stress} "
                      f"is negative, using alternative calculation")
            fatigue_safety_factor = endurance_limit / alt_eq_stress
            criterion = 'fatigue'

        static_safety_factor = FailureCriteria.langer_static_yield(yield_strength, alt_eq_stress,
                                                                   mean_eq_stress)
        if verbose:
            print(f"the {criterion} safety factor is: {fatigue_safety_factor}\n"
                  f"the Langer static safety factor is: {static_safety_factor}")

        return fatigue_safety_factor, static_safety_factor



"""module containing the EnduranceLimit class"""
from math import sqrt


class EnduranceLimit:
    """calculates Marin modification factors and return modified endurance limit"""

    def __init__(self, unmodified_Se, Sut, surface_finish, rotating, max_normal_stress,
                 max_bending_stress, stress_type, temp, reliability,
                 A95=None, diameter=None, height=None, width=None):
        """ Instantiating EnduranceLimit Object
        :param float unmodified_Se: The unmodified endurance strength
        :param float Sut: Ultimate tensile strength
        :param str surface_finish: 'ground' / 'machined' / 'cold-drawn' / 'hot-rolled' / 'as forged'
        :param bool rotating: rotating mode (True/False)
        :param float max_normal_stress: (for axial loading check)
        :param float max_bending_stress: (for axial loading check)
        :param str stress_type: 'bending' / 'axial' / 'torsion' / 'shear' / 'multiple'
        :param float temp: temperature
        :param float reliability: reliability
        :param float A95: Area containing over 95% of maximum periodic stress in the cross-section
        :param float diameter:
        :param float height:
        :param float width:
        """

        self.Sut = Sut
        self.surface_finish = surface_finish
        self.rotating = rotating
        self.max_normal_stress = max_normal_stress
        self.max_bending_stress = max_bending_stress
        self.diameter = diameter
        self.width = width
        self.height = height
        self.stress_type = stress_type
        self.temp = temp
        self.reliability = reliability
        self.unmodified = unmodified_Se
        self.A95 = A95

    @property
    def A95(self):
        return self._A95

    @A95.setter
    def A95(self, A95):
        if A95 is None:
            self._A95 = self.calc_A95()
        else:
            self._A95 = A95

    def calc_A95(self):
        if self.diameter is not None:
            return 0.01046 * self.diameter ** 2

        elif (self.width and self.height) is not None:
            return 0.05 * self.width * self.height

        else:
            raise ValueError('A95 is None and no parameters (diameter/width/height)'
                             'were given in order to calculate it')

    @property
    def Ka(self):
        """Returns Surface condition modification factor"""

        data = {'ground': (1.58, -0.085),
                'machined': (4.51, -0.265),
                'cold-drawn': (4.51, -0.265),
                'hot-rolled': (57.7, -0.718),
                'as forged': (272, -0.995)}
        a, b = data[self.surface_finish]
        return a * (self.Sut ** b)

    @property
    def Kb(self):  # FIXME: fix Kb its badly writen
        """Returns size modification factor"""
        if self.max_normal_stress > 0.85 * self.max_bending_stress:
            # if axial loading accrue
            return 1
        elif self.rotating and self.diameter is not None:
            # rotating and round
            de = self.diameter
        else:
            # not rotating or not round
            de = sqrt(self.A95 / 0.07658)

        if 2.79 <= de <= 51:
            return 1.24 * (de ** -0.107)
        elif 51 < de <= 254:
            return 1.51 * (de ** -0.157)

    @property
    def Kc(self):
        """Returns load modification factor"""
        types = {'bending': 1, 'axial': 0.85, 'torsion': 0.59, 'shear': 0.59, 'multiple': 1}
        return types[self.stress_type]

    @property
    def Kd(self):
        """Returns temperature modification factor"""
        return self.calc_kd(self.temp)

    @staticmethod
    def calc_kd(temp):
        """Calculate temperature modification factor"""
        import numpy as np
        percentage = np.array([20, 50, 100, 150, 200, 250, 300, 350, 400, 450, 500, 550, 600])
        reliability_factors = np.array([1, 1.01, 1.02, 1.025, 1.02, 1, 0.975, 0.943, 0.9, 0.843,
                                        0.768, 0.672, 0.549])
        return np.interp(temp, percentage, reliability_factors)

    @property
    def Ke(self):
        """Returns reliability factor"""
        return self.calc_ke(self.reliability)

    @staticmethod
    def calc_ke(reliability):
        """Calculates reliability factor"""
        import numpy as np
        percentage = np.array([50, 90, 95, 99, 99.9, 99.99, 99.999, 99.9999])
        reliability_factors = np.array([1, 0.897, 0.868, 0.814, 0.753, 0.702, 0.659, 0.620])
        return np.interp(reliability, percentage, reliability_factors)

    @property
    def Kf(self):
        """Miscellaneous effects factor"""
        return 1

    @staticmethod
    def unmodified_Se(Sut, material):
        """Returns the unmodified endurance strength limit based
        on the material (steel/iron/aluminium/copper alloy) and ultimate_tensile_strength

        :param float Sut: Ultimate Tensile Strength
        :param string material: (steel/iron/aluminium/copper alloy)
        """

        data = {'steel': {'divider': 1400, 'lesser': 0.5 * Sut, 'grater': 700},
                'iron': {'divider': 400, 'lesser': 0.4 * Sut, 'grater': 160},
                'aluminium': {'divider': 330, 'lesser': 0.4 * Sut, 'grater': 130},
                'copper alloy': {'divider': 280, 'lesser': 0.4 * Sut, 'grater': 100}}

        if Sut < data[material]['divider']:
            return data[material]['lesser']
        else:
            return data[material]['grater']

    @property
    def modified(self):
        """Returns the endurance limit modified
        by the Marin factors
        """
        return self.Ka * self.Kb * self.Kc * self.Kd * self.Ke * self.Kf * self.unmodified

    def get_factors(self, verbose=True):
        """Prints Marine factors
        :param bool verbose: Enables Marin factors printing

        :returns: Marin factors
        :rtype: tuple[float]
        """
        if verbose:
            print(f"Ka={self.Ka:.3f}, Kb={self.Kb:.3f}, Kc={self.Kc:.3f}, "
                  f"factor_Ks={self.Kd:.3f}, Ke={self.Ke:.3f}, Kf={self.Kf:.3f}")

        return self.Ka, self.Kb, self.Kc, self.Kd, self.Ke, self.Kf



"""A module containing the helical push spring class"""
from math import pi, sqrt
from sympy import sqrt

from me_toolbox.fatigue import FailureCriteria
from me_toolbox.springs import Spring
from me_toolbox.tools import percent_to_decimal


class HelicalCompressionSpring(Spring):
    """A helical push spring object"""
    def __repr__(self):
        return f"HelicalCompressionSpring(max_force={self.max_force}, " \
               f"wire_diameter={self.wire_diameter}, spring_diameter={self.diameter}, " \
               f"ultimate_tensile_strength={self.ultimate_tensile_strength}, " \
               f"shear_yield_percent={self.shear_yield_percent}, " \
               f"shear_modulus={self.shear_modulus}, elastic_modulus={self.elastic_modulus}, " \
               f"end_type={self.end_type}, spring_rate={self.spring_rate}, " \
               f"set_removed={self.set_removed}, shot_peened={self.shot_peened}, " \
               f"density={self.density}, working_frequency={self.working_frequency}, " \
               f"anchors={self.anchors}, zeta={self.zeta})"

    def __str__(self):
        return f"HelicalCompressionSpring(d={self.wire_diameter}, D={self.diameter}, " \
               f"k={self.spring_rate}, L0={self.free_length})"

    def __init__(self, max_force, wire_diameter, spring_diameter, ultimate_tensile_strength,
                 shear_yield_percent, shear_modulus, elastic_modulus, end_type,
                 spring_rate, set_removed=False, shot_peened=False,
                 density=None, working_frequency=None, anchors=None, zeta=0.15):
        """Instantiate helical push spring object with the given parameters.

        :param float max_force: The maximum load on the spring [N]
        :param float wire_diameter: Spring wire diameter [mm]
        :param float spring_diameter: Spring diameter measured from [mm]
            the center point of the wire diameter
        :param float ultimate_tensile_strength: Ultimate tensile strength of the material [MPa]
        :param float shear_yield_percent: Yield percent used to estimate shear_yield_stress
        :param float shear_modulus: Shear modulus [MPa]
        :param float or None elastic_modulus: Elastic modulus (used for buckling calculations) [MPa]
        :param str end_type: What kind of ending the spring has (effects length and number of coils)
            ,the options are: 'plain', 'plain and ground', 'squared or closed', 'squared and ground'
        :param float or None spring_rate: Spring rate (k) [N/mm]
        :param bool set_removed: If True adds to STATIC strength
            (must NOT use for fatigue application)
        :param bool shot_peened: If True adds to fatigue strength
        :param float or None density: Material density (used for finding natural frequency) [kg/m^3]
        :param float or None working_frequency: the spring working frequency [Hz]
        :param str or None anchors: How the spring is anchored (for buckling test),
            The options are: 'fixed-fixed', 'fixed-hinged', 'hinged-hinged', 'clamped-free'

            (used for buckling calculations)
        :param float zeta: Overrun safety factor

        :returns: HelicalCompressionSpring object
        """

        super().__init__(max_force, wire_diameter, spring_diameter, spring_rate,
                         ultimate_tensile_strength, shear_modulus, elastic_modulus,
                         shot_peened, density, working_frequency)

        if set_removed:
            print("Note: set should ONLY be removed for static loading"
                  "and NOT for periodical loading")

        self.set_removed = set_removed
        self.shear_yield_percent = shear_yield_percent
        self.zeta = zeta  # overrun safety factor
        self.end_type = end_type.lower()

        end_types = ('plain', 'plain and ground', 'squared or closed', 'squared and ground')
        if self.end_type not in end_types:
            raise ValueError(f"{end_type} not one of this: {end_types}")

        self.anchors = anchors

        self.check_design()

    def check_design(self):
        """Check if the spring index,active_coils,zeta and free_length
        are in acceptable range for good design

        :returns: True if all the checks are good
        :rtype: bool
        """
        good_design = True
        C = self.spring_index  # pylint: disable=invalid-name
        if isinstance(C, float) and not 4 <= C <= 12 and self.set_removed:
            print("Note: C - spring index should be in range of [4,12],"
                  "lower C causes surface cracks,\n"
                  "higher C causes the spring to tangle and requires separate packing")
            good_design = False
        elif isinstance(C, float) and not 3 <= C <= 12:
            print("Note: C - spring index should be in range of [3,12],"
                  "lower C causes surface cracks,\n"
                  "higher C causes the spring to tangle and requires separate packing")
            good_design = False

        active_coils = self.active_coils
        if isinstance(active_coils, float) and not 3 <= active_coils <= 15:
            print(f"Note: active_coils={active_coils:.2f} is not in range [3,15],"
                  f"this can cause non linear behavior")
            good_design = False

        zeta = self.zeta
        if zeta < 0.15:
            print(f"Note: zeta={zeta:.2f} is smaller then 0.15,"
                  f"the spring could reach its solid length")
            good_design = False
        if (self.free_length is not None) and (self.anchors is not None) \
                and (self.elastic_modulus is not None):
            buckling, safe_length = self.buckling(self.anchors, verbose=False)
            if buckling:
                print(f"Note: buckling is accruing, max free length"
                      f"(free_length)= {safe_length:.2f}, "
                      f"free_length= {self.free_length:.2f}")
                good_design = False

        if (self.density is not None) and (self.working_frequency is not None):
            natural_freq = self.natural_frequency
            if natural_freq <= 20 * self.working_frequency:
                print(
                    f"Note: the natural frequency={natural_freq:.2f} is less than 20*working"
                    f"frequency={20 * self.working_frequency:.2f} which is not good")
                good_design = False

        return good_design

    @property
    def free_length(self):
        """Calculates the free length of the spring"""
        return (self.Fsolid / self.spring_rate) + self.solid_length

    @property
    def solid_length(self):
        """Ls - the solid length of the spring
        (if the spring is fully compressed so the coils are touching each other)

        :returns: Spring solid length (when all the coils are touching)
        :rtype: float
        """
        diameter = self.wire_diameter
        total_coils = self.total_coils
        options = {'plain': diameter * (total_coils + 1),
                   'plain and ground': diameter * total_coils,
                   'squared or closed': diameter * (total_coils + 1),
                   'squared and ground': diameter * total_coils}
        return options.get(self.end_type)

    @property
    def Fsolid(self):  # pylint: disable=invalid-name
        """calculate the max_force necessary to bring the spring to solid length
        it is good practice for the max_force that compresses the spring to
        solid state to be greater than the maximum max_force anticipated, so we
        use this calculation: Fs=(1+zeta)Fmax in case the free length is unknown

        Note: zeta is the overrun safety factor, it's customary that zeta=0.15 so Fs=1.15Fmax

        :returns: The max_force it takes to get the spring to solid length
        :rtype: float
        """
        return (1 + self.zeta) * self.max_force

    @property
    def active_coils(self):
        """Calculate active_coils which is the number of active coils (using Castigliano's theorem)

        :returns: number of active coils
        :rtype: float
        """
        return ((self.shear_modulus * self.wire_diameter) /
                (8 * self.spring_index ** 3 * self.spring_rate)) * (
                       (2 * self.spring_index ** 2) / (1 + 2 * self.spring_index ** 2))

    @property
    def end_coils(self):
        """Ne - the end coils of the spring

        :returns: Number of the spring end coils
        :rtype: float
        """
        options = {'plain': 0,
                   'plain and ground': 1,
                   'squared or closed': 2,
                   'squared and ground': 2}
        return options.get(self.end_type)

    @property
    def total_coils(self):
        """Nt - the total coils of the spring

        :returns: Number of the spring total coils
        :rtype: float
        """
        return self.end_coils + self.active_coils

    @property
    def pitch(self):
        """ p - pitch of the spring (the distance between the coils)

        :returns: Pitch
        :rtype: float
        """
        options = {'plain': (self.free_length - self.wire_diameter) / self.active_coils,
                   'plain and ground': self.free_length / (self.active_coils + 1),
                   'squared or closed': ((self.free_length - 3 * self.wire_diameter) /
                                         self.active_coils),
                   'squared and ground': ((self.free_length - 2 * self.wire_diameter) /
                                          self.active_coils)}
        return options.get(self.end_type)

    @property
    def shear_yield_strength(self):
        """ Ssy - yield strength for shear
        (shear_yield_stress = % * ultimate_tensile_strength))

        :returns: yield strength for shear stress
        :rtype: float
        """
        try:
            return percent_to_decimal(self.shear_yield_percent) * self.ultimate_tensile_strength
        except TypeError:
            return self.shear_yield_percent * self.ultimate_tensile_strength

    @property
    def factor_Ks(self):  # pylint: disable=invalid-name
        """factor_Ks - Static shear stress concentration factor

        :returns: Static shear stress concentration factor
        :rtype: float
        """
        return (2 * self.spring_index + 1) / (2 * self.spring_index)

    @property
    def factor_Kw(self):  # pylint: disable=invalid-name
        """K_W - Wahl shear stress concentration factor

        :returns: Wahl shear stress concentration factor
        :rtype: float
        """
        return (4 * self.spring_index - 1) / (4 * self.spring_index - 4) + \
               (0.615 / self.spring_index)

    @property
    def factor_KB(self):  # pylint: disable=invalid-name
        """K_B - Bergstrasser shear stress concentration factor (very close to factor_Kw)

        NOTE:for the sake of completion NOT IMPLEMENTED!!!

        :returns: Bergstrasser shear stress concentration factor
        :rtype: float
        """
        return (4 * self.spring_index + 2) / (4 * self.spring_index - 3)

    @property
    def max_shear_stress(self):
        """ Return's the shear stress

        :returns: Shear stress
        :rtype: float
        """
        k_factor = self.factor_Ks if self.set_removed else self.factor_Kw
        return self.calc_shear_stress(self.max_force, k_factor)

    def calc_shear_stress(self, force, k_factor):
        """Calculates the max shear stress based on the max_force applied.

        :param float force: Working max_force of the spring
        :param float k_factor: the appropriate k factor for the calculation

        :returns: Shear stress
        :rtype: float
        """
        return (k_factor * 8 * force * self.diameter) / (pi * self.wire_diameter ** 3)

    @property
    def natural_frequency(self):
        """Figures out what is the natural frequency of the spring"""
        d = self.wire_diameter
        D = self.diameter
        Na = self.active_coils
        G = self.shear_modulus
        try:
            return (d * 1e-3 / (2 * D * 1e-3 ** 2 * Na * pi)) * sqrt(G / (2 * self.density))
        except TypeError:
            return None

    @property
    def max_deflection(self):
        """Returns the spring max_deflection, It's change in length

        :returns: Spring max_deflection
        :rtype: float
        """
        return self.calc_deflection(self.max_force)

    def calc_deflection(self, force):
        """Calculate the spring deflection (change in length) due to specific max_force.

        :param float force: Spring working max_force

        :returns: Spring deflection
        :rtype: float
        """
        C = self.spring_index
        d = self.wire_diameter
        G = self.shear_modulus
        Na = self.active_coils
        return ((8 * force * C ** 3 * Na) / (G * d)) * ((1 + 2 * C ** 2) / (2 * C ** 2))

    @property
    def weight(self):
        """Return's the spring *active coils* weight according to the specified density

        :returns: Spring weight
        :type: float
        """
        area = 0.25 * pi * (self.wire_diameter * 1e-3) ** 2  # cross-section area
        length = pi * self.diameter * 1e-3  # the circumference of the spring
        coil_volume = area * length
        if self.density is not None:
            return coil_volume * self.total_coils * self.density
        else:
            raise ValueError(f"Can't calculate weight, no density is specified")

    def static_safety_factor(self, solid=False):
        """ Returns the static safety factor according to the object attributes

        :returns: static factor of safety
        :type: float
        """
        k_factor = self.factor_Ks if self.set_removed else self.factor_Kw
        if solid:
            shear_stress = self.calc_shear_stress(self.Fsolid, k_factor)
        else:
            shear_stress = self.max_shear_stress
        return self.shear_yield_strength / shear_stress

    def fatigue_analysis(self, max_force, min_force, reliability,
                         criterion='modified goodman', verbose=False, metric=True):
        """ Returns safety factors for fatigue and for first cycle according to Lange failure
        criteria.

        :param float max_force: Maximal max_force acting on the spring
        :param float min_force: Minimal max_force acting on the spring
        :param float reliability: in percentage
        :param str criterion: fatigue criterion
        :param bool verbose: print more details
        :param bool metric: Metric or imperial

        :returns: static and dynamic safety factor
        :rtype: tuple[float, float]
        """
        # calculating mean and alternating forces
        alternating_force = abs(max_force - min_force) / 2
        mean_force = (max_force + min_force) / 2

        # calculating mean and alternating stresses
        k_factor = self.factor_Ks if self.set_removed else self.factor_Kw
        alt_shear_stress = self.calc_shear_stress(alternating_force, k_factor)
        mean_shear_stress = self.calc_shear_stress(mean_force, k_factor)

        Sse = self.shear_endurance_limit(reliability, metric)
        Ssu = self.shear_ultimate_strength
        Ssy = self.shear_yield_strength
        nf, nl = FailureCriteria.get_safety_factors(Ssy, Ssu, Sse, alt_shear_stress,
                                                    mean_shear_stress, criterion)
        if verbose:
            print(f"Alternating force = {alternating_force:.2f}, Mean force = {mean_force:.2f}\n"
                  f"Alternating shear stress = {alt_shear_stress:.2f}, "
                  f"Mean shear stress = {mean_shear_stress:.2f}\n"
                  f"Sse = {Sse:.2f}")
        return nf, nl

    def min_wire_diameter(self, Ap, m, safety_factor, diameter=None, spring_index=None,
                          solid=False):
        """The minimal wire diameter for given
        safety factor in order to avoid failure,
        according to the spring parameters.
        if solid is True the calculation uses :attr:`Fsolid`
        instead of :attr:`max_force`

        Note: In order for the calculation to succeed the spring
            diameter or the spring index should be known

        :param float Ap: for Sut calc
        :param float m: for Sut calc
        :param float safety_factor: Static safety factor
        :param float diameter: The spring diameter
        :param float spring_index: The spring index
        :param bool solid: If true calculate to according to the solid max_force

        :returns: The minimal wire diameter
        :rtype: float
        """
        if spring_index is not None:
            factor_ks = (2 * spring_index + 1) / (2 * spring_index)
            factor_kw = (4 * spring_index - 1) / (4 * spring_index - 4) + (0.615 / spring_index)
            factor_k = factor_ks if self.set_removed else factor_kw
            force = self.Fsolid if solid else self.max_force
            return ((8 * factor_k * force * spring_index * safety_factor) / (
                    self.shear_yield_percent * Ap * pi)) ** (1 / (2 - m))

        elif spring_index is None and diameter is not None:

            factor_k, temp_k = 1.1, 0
            diam = 0
            while abs(factor_k - temp_k) > 1e-4:
                # waiting for k to converge
                diam = ((8 * self.max_force * diameter * safety_factor * factor_k) / (
                        pi * self.shear_yield_percent * Ap)) ** (1 / (3 - m))
                temp_k = factor_k
                factor_ks = (2 * diameter + diam) / (2 * diameter)
                factor_kb = (4 * diameter + 2 * diam) / (4 * diameter - 3 * diam)
                factor_k = factor_ks if self.set_removed else factor_kb
            return diam
        else:
            print("Need to know spring index or spring diameter to calculate wire diameter")

    def min_spring_diameter(self, safety_factor, wire_diameter, solid=False):
        """return the minimum spring diameter to avoid static failure
        according to the specified safety factor, if the solid flag is True :attr:'Fsolid'
        is used instead of :attr:`max_force`.

        :param float safety_factor: static safety factor
        :param float wire_diameter: Spring's wire diameter
        :param bool solid: If true calculate to according to the solid max_force

        :returns: The minimal spring diameter
        :rtype: float
        """
        force = self.Fsolid if solid else self.max_force
        d = wire_diameter
        if self.set_removed:
            Ssy = self.shear_yield_strength
            return 0.5 * ((Ssy / safety_factor) * ((pi * d ** 3) / (4 * force)) - d)
        else:
            # derived using the KB factor (because it was easier)
            Sut = self.ultimate_tensile_strength
            alpha = (Sut * pi * d ** 3) / (8 * self.max_force * safety_factor)
            return 0.25 * ((2 * alpha - d) + sqrt((d - 2 * alpha) ** 2 - 24 * alpha * d))

    def buckling(self, anchors, verbose=True):
        """ Checks if the spring will buckle and find the
        maximum free length to avoid buckling
        :param str or None anchors: How the spring is anchored
            (The options are: 'fixed-fixed', 'fixed-hinged', 'hinged-hinged', 'clamped-free')
        :param bool verbose: Print buckling test result
        :returns: True if buckling occurring and The maximum safe length (free_length)
            to avoid buckling
        :rtype: tuple(bool, float)
        """
        # alpha values from table 10-2
        options = {'fixed-fixed': 0.5, 'fixed-hinged': 0.707, 'hinged-hinged': 1, 'clamped-free': 2}

        try:
            D = self.diameter
            E = self.elastic_modulus
            G = self.shear_modulus
            alpha = options[anchors.lower()]
            max_safe_length = (pi * D / alpha) * sqrt((2 * (E - G)) / (2 * G + E))
        except ValueError as err:
            print(f"{err}, make sure E and G have the same units (Mpa)")
        except KeyError as key:
            print(f"Ends: {key} is unknown ")
        except AttributeError:
            print("Anchors not specified")
        else:
            if verbose:
                if self.free_length >= max_safe_length:
                    print(f"Buckling is accruing, the max safe length = {max_safe_length:.2f}, ",
                          f"but the free_length = {self.free_length:.2f}")

                else:
                    print(
                        f"Buckling is NOT accruing, the max safe length = {max_safe_length:.2f}, ",
                        f"and the free_length = {self.free_length:.2f}")

            return self.free_length >= max_safe_length, max_safe_length

    @staticmethod
    def calc_spring_rate(wire_diameter, spring_diameter, shear_modulus, active_coils):
        """Calculate spring constant (using Castigliano's theorem)

        :returns: The spring constant
        :rtype: float
        """
        d = wire_diameter
        D = spring_diameter
        G = shear_modulus
        C = D / d
        Na = active_coils
        return ((G * d) / (8 * C ** 3 * Na)) * ((2 * C ** 2) / (1 + 2 * C ** 2))



"""module containing the ThreadedFastener class used for fastener strength analysis"""
from copy import deepcopy

from math import tan, radians, pi, log

from numpy import array

# from me_toolbox.fatigue import FatigueAnalysis
from me_toolbox.fasteners import Bolt
from me_toolbox.tools import print_atributes


class ThreadedFastener:
    # TODO: add pre-torque calculation
    def __repr__(self):
        return f"ThreadedFastener(bolt={self.bolt}, layers={self.layers}, nut={self.nut}, " \
               f"preload={self.preload})"

    def __str__(self):
        return f"Fastener(M{self.bolt.diameter})"

    def __init__(self, bolt, layers, nut, preload=None):
        """Initialize threaded fastener object
        :param Bolt bolt: A bolt object
        :param list[list] layers: lists of layers thicknesses and elastic modulus
        e.g. [[10,207e3], [5,70e3], [5,207e3]]
        :param bool nut: True if a nut is used, False if the last layer is threaded
        :param float or none preload: The initial load on the bolt (estimated if None)
        """

        self.bolt = bolt
        self.layers = layers
        self.nut = nut
        if preload is None:
            self.preload = bolt.estimate_preload(True)
            print(f"{self} - No preload was entered so an estimated value was used"
                  f"({round(self.preload,2)})"
                  f" under the assumption that the bolt is reusable ")
        else:
            self.preload = preload

    def get_info(self):
        """print all the fastener properties"""
        print_atributes(self)

    @property
    def grip_length(self):
        """griped length in the member (l)"""
        if self.nut:
            return sum([layer[0] for layer in self.layers])
        else:
            sum_of_unthreaded_layers = sum([layer[0] for layer in self.layers[:-1]])
            if self.layers[-1][0] < self.bolt.diameter:
                grip_length = sum_of_unthreaded_layers + 0.5 * self.layers[-1][0]
            else:
                grip_length = sum_of_unthreaded_layers + 0.5 * self.bolt.diameter
            return grip_length

    @property
    def griped_thread_length(self):
        """threaded section in grip (lt)"""
        lt = self.grip_length - self.bolt.shank_length
        if lt <= 0:
            raise ValueError(f"The bolt's shank length ({self.bolt.shank_length}) "
                             f"is larger than the griped length({self.grip_length})")
        return lt

    @property
    def bolt_stiffness(self):
        """bolt stiffness (Kb)"""
        bolt = self.bolt
        Ad = bolt.nominal_area
        At = bolt.stress_area
        E = bolt.elastic_modulus
        ld = bolt.shank_length
        lt = self.griped_thread_length
        # print(f"Ad={Ad},At={At},E={E},ld={ld},lt={lt},L={bolt.length},LT={bolt.thread_length},l={self.grip_length}")
        return (Ad * At * E) / ((Ad * lt) + (At * ld))

    @property
    def member_stiffness(self):
        """ member stiffness (Kb) """
        d = self.bolt.diameter
        D1 = self.bolt.head_diameter
        lt = self.grip_length

        return self.calc_member_stiffness(d, D1, lt, self.layers, self.nut)

    @staticmethod
    def calc_member_stiffness(diameter, head_diam, grip_length, Layers, nut=True, verbose=False):
        """Calculates member stiffness (Kb)
        :param float diameter: Bolt's nominal diameter
        :param float head_diam: Bolt's head diameter
        :param float grip_length: Length of gripped material
        :param list[list] Layers: tuple (or list)
            containing a tuple (or list) of layer thickness and material
        :param bool nut: True if fastener has a nut, False if the last layer is threaded
        :param bool verbose: print details for each layer

        :returns: Substrate stiffness
        :rtype: float
        """
        alpha = radians(30)  # angle of the approximated stress cone
        layers = deepcopy(Layers)

        if not nut:
            if layers[-1][0] < diameter:
                # replace the last layer for a layer half its size
                layers[-1][0] *= 0.5
            else:
                # replace the last layer for a layer half the nominal diameter size
                layers[-1][0] = 0.5 * diameter

        thicknesses = [layer[0] for layer in layers]
        elastic_modulus = [layer[1] for layer in layers]

        # finding the layer divided by the center line
        half_grip_len = 0.5 * grip_length
        tot = 0
        middle_index = 0
        for index, width in enumerate(thicknesses):
            tot += width
            if tot >= half_grip_len:
                middle_index = index
                break

        thickness_before_center_layer = sum(thicknesses[:middle_index])
        thickness_including_center_layer = tot

        if (thickness_including_center_layer - half_grip_len) != 0:
            # if half the grip length is not equal exactly
            # to the sum of layers composing it, split middle layer in to two parts
            thicknesses[middle_index] = (half_grip_len - thickness_before_center_layer)
            thicknesses.insert(middle_index + 1, thickness_including_center_layer - half_grip_len)
            elastic_modulus.insert(middle_index + 1, elastic_modulus[middle_index])

        diam = [head_diam]
        for index, thickness in enumerate(thicknesses):
            if index <= middle_index:
                diam.append(diam[index] + 2 * thickness * tan(alpha))
            else:
                diam.append(diam[index] - 2 * thickness * tan(alpha))

        diam.pop(middle_index + 1)  # removing the center line diameter

        stiffness = []
        d = diameter
        for D, t, E in zip(diam, thicknesses, elastic_modulus):
            ln = log(((1.155 * t + D - d) * (D + d)) / ((1.155 * t + D + d) * (D - d)))
            ki = (0.5774 * pi * E * d) / ln
            stiffness.append(ki)

            if verbose:
                print(f"d={d}, D={D}, t={t}, E={E}, ki={ki:.2f}")

        km_inv = sum(1 / array(stiffness))
        if verbose:
            print(f"Km={1 / km_inv:.2f}")
        return 1 / km_inv

    @property
    def fastener_stiffness(self):
        """Fastener stiffness of the joint (C),
        the fraction of external load carried by bolt
        """
        return self.bolt_stiffness / (self.member_stiffness + self.bolt_stiffness)

    def bolt_load(self, external_force):
        """The load on the bolt (Fb)"""
        return self.fastener_stiffness * external_force + self.preload

    def member_load(self, external_force):
        """The load on the member (Fm)"""
        return (1 - self.fastener_stiffness) * external_force - self.preload

    def safety_factors(self, external_force):
        """Safety factor for direct normal stress only
        (not shear stress and not eccentric loading)
        :param float external_force: The force acting on the fastener

        :return: A dictionary with the static safety factors<br>
                 n0 - Separation safety factor<br>
                 nL - Load safety factor<br>
                 np - Proof safety factor
        :rtype: dict
        """
        bolt_load = self.bolt_load(external_force)
        separation_safety_factor = self.preload / (external_force * (1 - self.fastener_stiffness))
        load_safety_factor = (self.bolt.proof_load - self.preload) / (bolt_load - self.preload)
        proof_safety_factor = self.bolt.proof_strength / bolt_load
        return {'n0': separation_safety_factor,
                'nL': load_safety_factor,
                'np': proof_safety_factor}



"""Module containing the Transmission Class"""
# I want the variables names to be the same as in AGMA pylint: disable=invalid-name
from math import cos, sin, log, sqrt, radians, pi

from me_toolbox.gears import Gear
from me_toolbox.tools import print_atributes


class GearTypeError(ValueError):
    """Error class, inherits from ValueError"""
    pass


class Transmission:
    """ Transmission object containing the transmission design parameters
        and methods to perform strength analysis on its gears. (AGMA 2001-D04)
    """

    # TODO: add advices to improve strength
    # TODO: add function to find minimum volume for contact and bending
    # TODO: add gear train calculation, including multiple stages

    def __init__(self, driving_machine, driven_machine, oil_temp, reliability,
                 power, SF, gear1, gear2=None, gear_ratio=0, SH=1):
        """
        :param Gear gear1: gear object (the driving gear by convention)
        :param Gear or None gear2: gear object (the driven gear by convention)
        :param str driving_machine: ('uniform' - Electric Motor/Turbine,
            'light shock' - multi-cylinder engine, 'medium shock' - single-cylinder engine)
        :param str driven_machine: (uniform / moderate shock / heavy shock)
        :param int oil_temp: oil temperature of the transmission in [deg spring_index]
        :param float reliability: reliability of the transmission
        :str float power: transmission power in [W]
        :param float SF: bending safety factor
        :keyword float SH: contact safety factor
        :keyword float gear_ratio: transmission gear ratio

        :returns: Transmission object
        :rtype: Transmission
        """
        if (type(gear1) is not type(gear2)) and (gear2 is not None):
            raise GearTypeError("gears entered to Transmission are not of the same type")

        self.gear1 = gear1
        self.gear2 = self._gear2_checkup(gear1, gear2, gear_ratio)

        # check if the pressure angle and modulus match between gear1 and gear2
        gear1.check_compatibility(self.gear2)

        self.driving_machine = driving_machine
        self.driven_machine = driven_machine
        self.oil_temp = oil_temp
        self.reliability = reliability
        self.power = power
        self.SF = SF
        self.SH = SH
        self.gear_ratio = gear_ratio
        self._contact_ratio()
        self.gear1.Y_j(self.gear1, self.gear2)
        self.gear1.Zw = 1  # the pinion's Zw is 1 by the AGMA standard
        self.gear2.Zw = self._Zw()
        self.ZI = gear1.ZI(self.gear1, self.gear2)

    def get_info(self):
        """print all the class fields with values"""
        print_atributes(self)

    def get_factors(self, verbose=True):
        """ retrieve transmission analysis factors

        :param bool verbose: print factors
        :type verbose: bool
        :rtype: dict
        """
        if verbose:
            print(f"Ko={self.Ko}, Yθ={self.Ytheta}, Yz=, {self.Yz}, ZE={self.ZE}, ZI={self.ZI}")

        return {"Ko=": self.Ko, "Yθ=": self.Ytheta, "Yz=": self.Yz, "ZE=": self.ZE, "ZI=": self.ZI}

    @property
    def Ytheta(self):
        """Returns temperature factor"""

        if self.oil_temp > 71:
            y_theta = (273 + self.oil_temp) / 344
        else:
            y_theta = 1
        return y_theta

    @property
    def Ko(self):
        """ Returns overload factor
        Ko is dependent on the type of driving motor type
        (Electric Motor/Turbine - uniform, multi-cylinder engine - light shock,
        single-cylinder engine - medium shock) and on the driven machine type
        (uniform, moderate shock, heavy shock)
        """
        table = {'uniform': {'uniform': 1, 'moderate shock': 1.25, 'heavy shock': 1.75},
                 'light shock': {'uniform': 1.25, 'moderate shock': 1.5, 'heavy shock': 2},
                 'medium shock': {'uniform': 15, 'moderate shock': 1.75, 'heavy shock': 2.25}}

        if self.driving_machine not in table:
            raise ValueError("error at Ko factor: invalid driving machine")

        if self.driven_machine not in table[self.driving_machine]:
            raise ValueError("error at Ko factor: invalid driven machine")

        return table[self.driving_machine][self.driven_machine]

    @property
    def Yz(self):
        """Returns reliability factor"""
        R = self.reliability
        if 0.9 <= R <= 0.99:
            Y_z = 0.658 - 0.0759 * log(1 - R)
        elif 0.99 < R <= 0.9999:
            Y_z = 0.5 - 0.109 * log(1 - R)
        else:
            raise ValueError("Reliability not in range (0.9<=R<=0.9999)")

        return Y_z

    @property
    def ZE(self):
        """returns the elastic coefficient"""
        elastic_modulus_list = {'steel': 2e5, 'malleable iron': 1.7e5,
                                'nodular iron': 1.7e5, 'cast iron': 1.5e5,
                                'aluminum bronze': 1.2e5, 'tin bronze': 1.1e5}

        material1 = self.gear1.material
        material2 = self.gear2.material

        E1 = elastic_modulus_list.get(material1, material1)
        E2 = elastic_modulus_list.get(material2, material2)

        poissons_ratio = 1 / 3
        try:
            return sqrt(
                (1 / pi) / (((1 - poissons_ratio ** 2) / E1) + ((1 - poissons_ratio ** 2) / E2)))
        except TypeError:
            print(f"error: at ZE: invalid gear material ({material1} or {material2})")

    @property
    def centers_distance(self):
        """Returns the distance between the gear's centers"""
        return self.gear1.calc_centers_distance(self.gear_ratio)

    # bending stress related methods
    def bending_stress(self, gear):
        """Calculating bending stress

        :param Gear gear: gear object

        :returns: Bending stress
        :rtype: float
        """
        N = gear.teeth_num
        b = gear.width
        d = gear.pitch_diameter
        Yj = gear.Yj
        Wt = gear.calc_forces(gear, self.power)[0]
        sigma = (Wt * N * self.Ko * gear.Kv * gear.Ks * gear.KH * gear.KB) / (
                    Yj * b * d)
        return sigma

    def allowed_bending_stress(self, gear):
        """Calculating allowed bending stress

        :para Gear gear: gear object

        :returns: Allowed bending stress
        :rtype: float
        """
        allowed_bending_stress = (gear.St * gear.YN) / (self.Ytheta * self.Yz * self.SF)
        return allowed_bending_stress

    def minimum_width_for_bending(self, gear):
        """Calculating minimum gear width to withstand bending stress

        :keyword Gear gear: gear object

        :returns: Minimum gear width
        :rtype: float
        """
        Yj = gear.Yj
        N = gear.teeth_num
        d = gear.pitch_diameter

        allowed_bending = self.allowed_bending_stress(gear)
        Wt = gear.calc_forces(gear, self.power)[0]
        minimum_gear_width = ((Wt * N * self.Ko * gear.Kv * gear.Ks * gear.KH *
                               gear.KB) / (Yj * allowed_bending * d))
        return minimum_gear_width

    # contact stress related methods
    def contact_stress(self, gear):
        """Calculating contact stress

        :param Gear gear: gear object

        :returns: Contact stress
        :rtype: float
        """
        b = gear.width
        d = gear.pitch_diameter
        Wt = gear.calc_forces(gear, self.power)[0]

        sigma = sqrt((Wt * (self.ZE ** 2) * self.Ko * gear.Kv * gear.Ks * gear.KH * gear.ZR)
                     / (b * d * self.ZI))
        return sigma

    def allowed_contact_stress(self, gear):
        """Calculating allowed contact stress

        :param Gear gear: gear object

        :returns: Allowed contact stress
        :rtype: float
        """
        allowed_contact_stress = (gear.Sc * gear.ZN * gear.Zw) / (self.Ytheta * self.Yz * self.SH)
        return allowed_contact_stress

    def minimum_width_for_contact(self, gear):
        """Calculating minimum gear width to withstand contact stress

        :param Gear gear: gear object

        :returns: Minimum gear width
        :rtype: float
        """
        Wt = gear.calc_forces(gear, self.power)[0]
        d = gear.pitch_diameter
        allowed_contact = self.allowed_contact_stress(gear)
        Kv = gear.Kv
        Ks = gear.Ks
        KH = gear.KH
        ZR = gear.ZR

        minimum_gear_width = ((Wt * self.ZE ** 2 * self.Ko * Kv * Ks * KH * ZR) /
                              (d * self.ZI * allowed_contact ** 2))
        return minimum_gear_width

    # for both bending and contact stresses
    def life_expectency(self, gear, in_hours=False):
        """Calculates expected life span of the gear,
         if not possible return  stress cycle factors (YN/ZN)

        :example: helical = HelicalGear(gear_properties)
                  gearbox = Transmission(transmission_properties)
                  gearbox.LifeExpectency(helical, in_hours=True)

        :param Gear gear: gear object
        :param bool in_hours: return gears life expectency in hours

        :returns: Number of cycles or work hours
        :rtype: float
        """

        bending_stress = self.bending_stress(gear)
        contact_stress = self.contact_stress(gear)
        YN = (bending_stress * self.SF * self.Ytheta * self.Yz) / gear.St
        ZN = (contact_stress * self.SH * self.Ytheta * self.Yz) / (gear.Sc * gear.Zw)

        # for YN > 1
        YN_low_cycle = {160: (YN / 2.3194) ** (-1 / 0.0538),
                        250: (YN / 4.9404) ** (-1 / 0.1045),
                        400: (YN / 9.4518) ** (-1 / 0.148)}
        # for YN <= 1
        YN_high_cycle = {True: (YN / 1.6831) ** (-1 / 0.0323),
                         False: (YN / 1.3558) ** (-1 / 0.0178)}

        # for ZN > 1
        ZN_low_cycle = {True: (ZN / 1.249) ** (-1 / 0.0138),
                        False: (ZN / 2.466) ** (-1 / 0.056)}
        # for ZN < 1
        ZN_high_cycle = {True: (ZN / 2.466) ** (-1 / 0.056),
                         False: (ZN / 1.4488) ** (-1 / 0.023)}
        # print(f"YN = {YN}, ZN = {ZN}")
        try:
            # number of cycles until bending failure
            Ny = YN_low_cycle[gear.hardness] if YN > 1 else YN_high_cycle[gear.sensitive_use]

            # number of cycles until contact failure
            Nz = ZN_low_cycle[gear.nitriding] if ZN > 1 else ZN_high_cycle[gear.sensitive_use]

            N = min(Ny, Nz)
            # print(f"Ny={Ny:e}, Nz={Nz:e}")
        except KeyError:
            print(f"error: YN > 1 but hardness {gear.hardness} has no graph associated with it")
            # return YN, ZN
        else:
            # if in_hours True convert number of cycles to house
            # and shorten float length to only 2 decimal places
            return float(f"{N / (gear.rpm * 60):.2f}") if in_hours else N

    def minimal_hardness(self, gear):
        """Returns the minimal hardness of the gear to avoid failure

        :param Gear gear: Gear object

        :returns: Minimal hardness
        :rtype: float
        """
        # for bending
        St = (self.Ytheta * self.Yz * self.SF * self.bending_stress(gear)) / gear.YN
        if gear.grade == 1:
            # for grade 1
            HBt = (St - 88.3) / 0.533
        else:
            # for grade 2
            HBt = (St - 113) / 0.703

        # for contact
        Sc = (self.Ytheta * self.Yz * self.SH * self.bending_stress(gear)) / (gear.ZN * gear.Zw)
        if gear.grade == 1:
            # for grade 1
            HBc = (Sc - 200) / 2.22
        else:
            # for grade 2
            HBc = (Sc - 237) / 2.41

        return max(HBt, HBc)

    def optimize(self, gear, optimize_feature='all', verbose=False):
        """ perform gear optimization

        example: result, results_list = gearbox.Optimize(pinion, optimize_feature='width')

        note: result of width in [mm], volume in [mm^3] and center distance in [mm]

        :param Gear gear: gear object
        :param str optimize_feature: property to optimize for ('width'/'volume'/'center')
        :param bool verbose: print optimization stages

        :returns: An optimized result and list of other viable options
        :rtype: tuple
        """
        return gear.optimization(self, optimize_feature, verbose)

    def check_undercut(self):
        """Checks undercut state """

        phi = radians(self.gear1.pressure_angle)
        invert_mG = 1 / self.gear_ratio
        minimum_teeth_num = 2 * (1 + sqrt(1 + invert_mG * (2 + invert_mG) * sin(phi) ** 2)) / (
                (2 + invert_mG) * sin(phi) ** 2)
        if self.gear1.teeth_num > minimum_teeth_num:
            print("No Undercut")
        else:
            print("Undercut Occurs")

    # for internal use
    def _Zw(self):
        """Surface strength factor"""
        if self.gear1 is not None:
            HBp = self.gear1.hardness
        else:
            HBp = self.gear2.hardness
        if self.gear2 is not None:
            HBg = self.gear2.hardness
        else:
            HBg = self.gear1.hardness

        ratio = HBp / HBg

        if 1.2 <= ratio < 1.7:
            A = 0.00898 * ratio - 0.00829
        elif ratio >= 1.7:
            A = 0.00698
        else:
            # if ratio < 1.2:
            A = 0

        return 1 + A * (self.gear_ratio - 1)

    def _contact_ratio(self):
        """Calculate contact ratio between the gears
        and assign it to the gear's attributes
        """
        rp = 0.5 * self.gear1.pitch_diameter
        # check if gear2 is specified
        if self.gear2 is not None:
            rG = 0.5 * self.gear2.pitch_diameter
        else:
            rG = rp * self.gear_ratio
        phi = radians(self.gear1.pressure_angle)
        m = self.gear1.modulus
        p = self.gear1.pitch
        contact_length = (sqrt((rG + m) ** 2 - (rG * cos(phi)) ** 2) +
                          sqrt((rp + m) ** 2 - (rp * cos(phi)) ** 2) - (rp + rG) * sin(phi))
        contact_ratio = contact_length / (p * cos(phi))
        if contact_ratio < 1.2:
            print("attention: ratio is should be higher than 1.2")

        self.gear1.contact_ratio = contact_ratio
        self.gear2.contact_ratio = contact_ratio

    @staticmethod
    def _gear2_checkup(gear1, gear2, gear_ratio):
        """Instantiate gear2 object if it wasn't given

        :param Gear gear1: gear object
        :param Gear or None gear2: gear object
        :param float gear_ratio: gear ratio

        :returns: Gear2 object
        :rtype: Gear
        """

        if gear2 is None and gear_ratio != 0:
            # create gear2 from gear1 properties and the gear ratio

            # copy gear1 properties
            gear2_prop = gear1.__dict__.copy()
            # change the name, teeth number and rpm according to the gear ratio
            gear2_prop.update(
                {'teeth_num': round(gear1.teeth_num * gear_ratio), 'rpm': gear1.rpm / gear_ratio})

            return gear1.create_new_gear(gear1.format_properties(gear2_prop))

        elif gear2 is not None and gear_ratio == 0:
            # gear2 had been specified
            return gear2
        elif gear2 is not None and gear_ratio != 0:
            # both gear2 and gear ratio entered check if they match
            if gear2.teeth_num / gear1.teeth_num == gear_ratio:
                return gear2
            else:
                # mismatch gear ratio
                raise GearTypeError(f"calculated gear ratio:{gear2.teeth_num / gear1.teeth_num}"
                                    f" not matching specified gear ratio:{gear_ratio}")
        else:
            # gear2 cant be None at the same time gear_ratio is zero
            raise GearTypeError('gear2 and gear_ratio not specified, one of them is needed')

    @staticmethod
    def train_value(*gear_pairs):
        """Returns transmission train value
        (which is the inverse of the transmission value)
        the input should be an iterable of two gears
        each of them can be either int or Gear obj

        Note: A negative sign indicates direction change of rotation

        Example:
            >> N1, N2, N3, N4 = 10, 20, 50, 10
            >> e = gearbox.train_value((N1, N2), (N3, N4)))

        :param gear_pairs: iterable of two gears
        """
        t_val = 1
        for gears in gear_pairs:
            if hasattr(gears, '__iter__'):
                N1 = gears[0].teeth_num if isinstance(gears[0], Gear) else gears[0]
                N2 = gears[1].teeth_num if isinstance(gears[1], Gear) else gears[1]
                t_val *= -(N1 / N2)
            else:
                raise TypeError("Gear pair can only be iterables")
        return t_val



from unittest import TestCase
from math import log, pi, tan, radians
import numpy as np

from me_toolbox.fasteners import Bolt
from me_toolbox.fasteners import ThreadedFastener


def stiffness(d, Diam, thickness, Elastic):
    stiff = []
    for D, t, E in zip(Diam, thickness, Elastic):
        ln = log(((1.155 * t + D - d) * (D + d)) / ((1.155 * t + D + d) * (D - d)))
        stiff.append(0.5774 * pi * E * d / ln)
    km_inv = sum(1 / np.array(stiff))
    return 1 / km_inv


class TestThreadedFastener(TestCase):
    tan_alpha = tan(radians(30))

    def setUp(self):
        diameter = 10
        pitch = 1.5
        length = 20
        threaded_length = 15
        grade = '12.9'
        E = 207e3

        Sp, Sut, Sy = Bolt.get_strength_prop(diameter, grade)

        self.bolt = Bolt(diameter, pitch, length, threaded_length, Sy, Sut, Sp, E)

        thickness = [2, 3, 3, 2, 3]
        Elastic = [200e3, 70e3, 200e3, 200e3, 200e3]
        self.layers = [[t, E] for t, E in zip(thickness, Elastic)]
        self.grip_len = (sum([layer[0] for layer in self.layers]))
        self.fastener = ThreadedFastener(self.bolt, self.layers, nut=True)

    def test_griped_thread(self):
        self.assertAlmostEqual(self.fastener.griped_thread_length, 8)

    def test_bolt_stiffness(self):
        self.assertAlmostEqual(self.fastener.bolt_stiffness, 1026695.2877576174)

    def test_substrate_stiffness_1(self):
        # test if the first substrate is smaller than half the material
        thickness = [2, 3, 3, 2, 3]
        Elastic = [200e3, 70e3, 200e3, 200e3, 200e3]
        layers = [(t, E) for t, E in zip(thickness, Elastic)]

        D1, D6 = 15, 15
        D2 = D1 + 2 * thickness[0] * self.tan_alpha
        D3 = D2 + 2 * thickness[1] * self.tan_alpha
        D5 = D6 + 2 * thickness[2] * self.tan_alpha
        D4 = D5 + 2 * thickness[3] * self.tan_alpha
        Diam = [D1, D2, D3, D4, D5, D6]
        km = stiffness(self.bolt.diameter, Diam, [2, 3, 1.5, 1.5, 2, 3],
                       [200e3, 70e3, 200e3, 200e3, 200e3, 200e3])
        self.fastener.layers = layers
        self.assertAlmostEqual(self.fastener.substrate_stiffness, km)

    def test_substrate_stiffness_2(self):
        # test if the first substrate is smaller than half the material
        thickness = [2, 4, 4]
        Elastic = [200e3, 70e3, 200e3]
        layers = [(t, E) for t, E in zip(thickness, Elastic)]
        D1, D4 = 15, 15
        D2 = D1 + 2 * thickness[0] * self.tan_alpha
        D3 = D4 + 2 * thickness[2] * self.tan_alpha
        Diam = [D1, D2, D3, D4]

        km = stiffness(self.bolt.diameter, Diam, [2, 3, 1, 4], [200e3, 70e3, 70e3, 200e3])
        self.fastener.layers = layers
        self.assertAlmostEqual(self.fastener.substrate_stiffness, km)

    def test_substrate_stiffness_3(self):
        # test if the first substrate is larger than half the material
        thickness = [5, 1, 2]
        Elastic = [200e3, 70e3, 200e3]
        layers = [(t, E) for t, E in zip(thickness, Elastic)]
        D1, D4 = 15, 15
        D3 = D4 + 2 * thickness[2] * self.tan_alpha
        D2 = D3 + 2 * thickness[1] * self.tan_alpha
        Diam = [D1, D2, D3, D4]
        km = stiffness(self.bolt.diameter, Diam, [4, 1, 1, 2], [200e3, 200e3, 70e3, 200e3])
        self.fastener.layers = layers
        self.assertAlmostEqual(self.fastener.substrate_stiffness, km)

    def test_substrate_stiffness_4(self):
        # test symmetric layers
        thickness = [2, 4, 2]
        Elastic = [200e3, 70e3, 200e3]
        layers = [(t, E) for t, E in zip(thickness, Elastic)]
        D1, D4 = 15, 15
        D2 = D1 + 2 * thickness[0] * self.tan_alpha
        D3 = D4 + 2 * thickness[2] * self.tan_alpha
        Diam = [D1, D2, D3, D4]
        km = stiffness(self.bolt.diameter, Diam, [2, 2, 2, 2], [200e3, 70e3, 70e3, 200e3])
        self.fastener.layers = layers
        self.assertAlmostEqual(self.fastener.substrate_stiffness, km)

    def test_substrate_stiffness_5(self):
        # test center between layers
        thickness = [2, 3, 3, 2]
        Elastic = [200e3, 70e3, 200e3, 200e3]
        layers = [(t, E) for t, E in zip(thickness, Elastic)]
        D1, D4 = 15, 15
        D2 = D1 + 2 * 2 * self.tan_alpha
        D3 = D4 + 2 * 2 * self.tan_alpha
        Diam = [D1, D2, D3, D4]
        km = stiffness(self.bolt.diameter, Diam, thickness, Elastic)
        self.fastener.layers = layers
        self.assertAlmostEqual(self.fastener.substrate_stiffness, km)

    def test_fastener_stiffness(self):
        self.assertAlmostEqual(self.fastener.fastener_stiffness, 0.33840496783181806)



from me_toolbox.gears.gear import Gear
from me_toolbox.gears.spur_gear import SpurGear
from me_toolbox.gears.helical_gear import HelicalGear
from me_toolbox.gears.transmission import Transmission, GearTypeError



"""module containing the table interpolation function"""
import numpy as np


class NotInRangeError(ValueError):
    def __init__(self, var, num, range_):
        """Error for when value is not in the specified range

        :param str var: var assigned to
        :param float num: num to assign
        :param tuple range_: permitted range
        """

        self.num = num
        self.range_ = range_
        self.msg = f"{var} = {num} not in range {range_}"
        super().__init__(self.msg)


def table_interpolation(x_row, x_col, data):
    """ Get table in numpy array form and two coordinates and
    Interpolate the value in the table corresponding to those coordinates

    :keyword x_row: the x row from which to retrieve the value
    :type x_row: float
    :keyword x_col: the x col from which to retrieve the value
    :type x_col: float
    :keyword data: the table as numpy array
    :type data: np.ndarray
    :rtype: float
    """

    # first gear number of known teeth values
    first_column = data[:, 0]

    # second gear number of known teeth values
    first_row = data[0, :]

    if x_col > first_row[-1] or x_col < first_row[1]:
        raise NotInRangeError("x_col", x_col, (first_row[1], first_row[-1]))

    if x_row > first_column[-1] or x_row < first_column[1]:
        raise NotInRangeError("x_row", x_row, (first_column[1], first_column[-1]))

    # upper bounds indexes of x_row and x_col
    index1 = np.searchsorted(first_column, x_row, side='right')
    index2 = np.searchsorted(first_row, x_col, side='right')

    try:
        # get x_row neighbors
        x_row_low, x_row_high = data[index1 - 1, 0], data[index1, 0]
    except IndexError:
        # if x_row upper neighbor is out of the table range
        x_row_low, x_row_high = data[index1 - 1, 0], np.nan
    try:
        # get x_col neighbors
        x_col_low, x_col_high = data[0, index2 - 1], data[0, index2]
    except IndexError:
        # if x_col upper neighbor is out of the table range
        x_col_low, x_col_high = data[0, index2 - 1], np.nan

    # if gear teeth number is not in the table do an
    # interpolation on closest values inside the table
    if x_row == x_row_low and x_col == x_col_low:
        # x_row and x_col are known no need to interpolate
        result = data[index1 - 1, index2 - 1]
    elif x_row == x_row_low:
        # x_row is known and x_col is interpolated
        result = np.interp(x_col, data[0, :], data[index1 - 1, :])
    elif x_col == x_col_low:
        # x_col is known and x_row is interpolated
        result = np.interp(x_row, data[:, 0], data[:, index2 - 1])
    else:
        # both are unknown and interpolated
        lower_bound = np.interp(x_row, data[:, 0], data[:, index2 - 1])
        upper_bound = np.interp(x_row, data[:, 0], data[:, index2])
        range_ = np.array([data[0, index2 - 1], data[0, index2]])
        bounds = np.array([lower_bound, upper_bound])
        result = np.interp(x_col, range_, bounds)

    return result



"""A module containing the spring class"""
import csv
import os
import numpy as np
from sympy import Symbol

from me_toolbox.tools import print_atributes


class Spring:

    def __repr__(self):
        return f"Spring(max_force={self.max_force}, wire_diameter={self.wire_diameter}, " \
               f"diameter={self.diameter}, spring_rate={self.spring_rate}, " \
               f"active_coils={self.active_coils}, " \
               f"ultimate_tensile_strength={self.ultimate_tensile_strength}, " \
               f"shear_modulus={self.shear_modulus}, elastic_modulus={self.elastic_modulus}," \
               f"shot_peened={self.shot_peened}, density={self.density}, " \
               f"working_frequency={self.working_frequency})"

    def __str__(self):
        return f"{self.__class__.__name__}(K={self.spring_constant}, d={self.wire_diameter}, " \
               f"D={self.diameter})"

    def __init__(self, max_force, wire_diameter, diameter, spring_rate,
                 ultimate_tensile_strength, shear_modulus, elastic_modulus, shot_peened,
                 density, working_frequency):
        """Instantiate helical push spring object with the given parameters
        :param float or Symbol max_force: The maximum load on the spring [N]
        :param float or Symbol wire_diameter: Spring wire diameter [mm]
        :param float or Symbol diameter: Spring diameter measured from [mm]
            the center point of the wire diameter
        :param float or None spring_rate: Spring rate (k) [N/mm]
        :param float ultimate_tensile_strength: Ultimate tensile strength of the material [MPa]
        :param float shear_modulus: Shear modulus [MPa]
        :param float or None elastic_modulus: Elastic modulus (used for buckling calculations) [MPa]
        :param bool shot_peened: If True adds to fatigue strength
        :param float or None density: Spring's material density [kg/m^3]
        :param float or None working_frequency: the spring working frequency [Hz]

        :returns: Spring object
        """

        self.max_force = max_force
        self.working_frequency = working_frequency
        self.wire_diameter = wire_diameter
        self.diameter = diameter
        self.spring_rate = spring_rate
        self.ultimate_tensile_strength = ultimate_tensile_strength
        self.shear_modulus = shear_modulus
        self.elastic_modulus = elastic_modulus
        self.density = density
        self.shot_peened = shot_peened

    def get_info(self):
        """print all the spring properties"""
        print_atributes(self)

    @property
    def wire_diameter(self):
        """Getter for the wire diameter attribute

        :returns: The spring's wire diameter
        :rtype: float or Symbol
        """
        return self._wire_diameter

    @wire_diameter.setter
    def wire_diameter(self, diameter):
        """Sets the wire diameter and updates relevant attributes
        :param float diameter: Spring's wire diameter
        """
        self._wire_diameter = diameter

    @property
    def diameter(self):
        """Getter for the spring diameter attribute

        :returns: The spring diameter
        :rtype: float or Symbol
        """
        return self._diameter

    @diameter.setter
    def diameter(self, diameter):
        """Sets the spring diameter and updates relevant attributes
        :param float diameter: Spring's diameter
        """
        self._diameter = diameter

    @property
    def inside_diameter(self):
        return self.diameter - self.wire_diameter

    @property
    def outside_diameter(self):
        return self.diameter + self.wire_diameter

    @property
    def spring_index(self):
        """C - spring index

        Note: C should be in range of [4,12], lower C causes surface cracks,
            higher C causes the spring to tangle and require separate packing

        :returns: The spring index
        :type: float or Symbol
        """
        return self.diameter / self.wire_diameter

    @property
    def spring_rate(self):
        """getter for the :attr:`spring_rate` attribute

        :returns: The spring rate
        :rtype: float
        """
        return self._spring_rate

    @spring_rate.setter
    def spring_rate(self, spring_rate):
        """setter for the :attr:`spring_rate` attribute
        :param float or None spring_rate: K - The spring rate
        """
        self._spring_rate = spring_rate

    @property
    def shear_ultimate_strength(self):
        """ Ssu - ultimate tensile strength for shear """
        return 0.67 * self.ultimate_tensile_strength

    def shear_endurance_limit(self, reliability, metric=True):
        """Sse - Shear endurance limit according to Zimmerli
        :param float reliability: reliability in percentage
        :param bool metric: metric or imperial

        :returns: Sse - Shear endurance limit
        :rtype: float
        """
        # data from table
        percentage = np.array([50, 90, 95, 99, 99.9, 99.99, 99.999, 99.9999])
        reliability_factors = np.array([1, 0.897, 0.868, 0.814, 0.753, 0.702, 0.659, 0.620])
        # interpolating from data
        Ke = np.interp(reliability, percentage, reliability_factors)

        if self.shot_peened:
            Ssa, Ssm = (398, 534) if metric else (57.5e3, 77.5e3)
        else:
            Ssa, Ssm = (241, 379) if metric else (35e3, 55e3)

        return Ke * (Ssa / (1 - (Ssm / self.shear_ultimate_strength) ** 2))

    @staticmethod
    def material_prop(material, diameter, metric=True, verbose=False):
        """Reads table A_and_m.csv from file and returns the Sut estimation from the material
         properties Ap and m
        :param str material: The spring's material
        :param float diameter: Wire diameter
        :param bool metric: Metric or imperial
        :param bool verbose: Prints Values of A and m

        :returns: ultimate tensile strength (Sut)
        :rtype: float
        """

        if isinstance(diameter, Symbol):
            raise ValueError(f"the material keyword can't be used if the diameter is symbolic "
                             f"specify Ap and m manually")

        path = os.path.dirname(__file__) + "\\tables\\A_and_m.csv"
        with open(path, newline='') as file:
            reader = csv.DictReader(file)
            table = []
            available_types = []
            for line in reader:
                table.append(line)
                available_types.append(line['type'])

        for line in table:
            min_d = float(line['min_d_mm'] if metric else line['min_d_in'])
            max_d = float(line['max_d_mm'] if metric else line['max_d_in'])
            if line['type'] == material.lower() and min_d <= diameter <= max_d:
                A, m = float(line['A_mm'] if metric else line['A_in']), float(line['m'])
                if verbose:
                    print(f"A={A}, m={m}")
                return A / (diameter ** m)

        if material not in available_types:
            raise KeyError("The material is unknown")
        else:
            raise ValueError("The diameter don't match any of the values in the table")



"""module containing the BoltPattern class used for a bolt pattern strength analysis"""
from numpy import array, cross, dot, sqrt
from numpy.linalg import norm

from me_toolbox.fatigue import FatigueAnalysis, EnduranceLimit
from me_toolbox.tools import print_atributes


class BoltPattern:
    def __repr__(self):
        return f"BoltPattern(fasteners={[str(fastener) for fastener in self.fasteners]}, " \
               f"fasteners_locations={self.fasteners_locations}, " \
               f"force={self.force}, force_location={self.force_location}, " \
               f"axis_of_rotation={self.axis_of_rotation}, shear_location={self.shear_location})"

    def __str__(self):
        return f"Pattern({[str(fastener) for fastener in self.fasteners]})"

    def __init__(self, fasteners, fasteners_locations, force, force_location,
                 axis_of_rotation, shear_location):
        """Initialize bolt pattern
        :param list[ThreadedFastener] fasteners: A List of threaded fasteners object
        e.g. [M10_fastener, M10_fastener, M8_fastener, M8_fastener]
        :param list[list] fasteners_locations: A list of coordinates for each of the
         fasteners locations
        :param list force: The external force in vector form [x,y,z]
        :param list force_location: The external force location in vector form [x,y,z]
        :param list[list] axis_of_rotation: List of two points that describe the axis of
        rotation
        :param string shear_location: Where along the volt the shear is felt, shank or thread
        used to determine what area value to use
        """

        self.fasteners = fasteners
        self.fasteners_locations = fasteners_locations
        self.force = force
        self.preloads = [load.preload for load in self.fasteners]
        self.force_location = force_location
        self.axis_of_rotation = axis_of_rotation
        self.shear_location = shear_location

    def get_info(self):
        """print all the fastener properties"""
        print_atributes(self)

    @property
    def fasteners_stiffness(self):
        """bolts' fastener stiffness (C)"""
        return [fastener.fastener_stiffness for fastener in self.fasteners]

    @property
    def total_stiffness(self):
        """fasteners' total stiffness list (Kb+Km)"""
        return [fastener.member_stiffness + fastener.bolt_stiffness for fastener in self.fasteners]

    @property
    def bolt_shear_area(self):
        """if the shear stress is in the threaded section the shear area is the stress areas
           if the shear stress is in the shank section the shear area is the shank area"""
        if self.shear_location == 'shank':
            return [fastener.bolt.nominal_area for fastener in self.fasteners]
        elif self.shear_location == 'thread':
            return [fastener.bolt.stress_area for fastener in self.fasteners]
        else:
            raise ValueError("shear_location can be 'shank' or 'thread'")

    @property
    def shear_stress(self):
        """bolts' shear stress"""
        return [norm(i) / j for i, j in zip(self.total_shear_force, self.bolt_shear_area)]

    @property
    def total_shear_force(self):
        """the total shear force on bolt from the direct shear force and resulting torque
        (Fi = Fvi+FGi)"""
        return [i + j for i, j in zip(self.direct_shear_force, self.torque_shear_force)]

    @property
    def direct_shear_force(self):
        """shear forces directly from the external force (Fvi)"""
        force = array(self.force)

        # zeroing forces in the z direction (because they don't cause shear)
        force[2] = 0

        return [(force * area) / sum(self.bolt_shear_area) for area in self.bolt_shear_area]

    @property
    def center_of_rotation(self):
        """center of rotation for an off center shear force"""
        bolts_x_locations = [bolt[0] for bolt in self.fasteners_locations]
        bolts_y_locations = [bolt[1] for bolt in self.fasteners_locations]
        bolts_z_locations = [bolt[2] for bolt in self.fasteners_locations]
        # calculating center of rotation (G)
        Gx, Gy, Gz = 0, 0, 0
        for area, location in zip(self.bolt_shear_area, bolts_x_locations):
            Gx += (area * location) / sum(self.bolt_shear_area)
        for area, location in zip(self.bolt_shear_area, bolts_y_locations):
            Gy += (area * location) / sum(self.bolt_shear_area)
        for area, location in zip(self.bolt_shear_area, bolts_z_locations):
            Gz += (area * location) / sum(self.bolt_shear_area)
        return Gx, Gy, Gz

    @property
    def torque_shear_force(self):
        """shear forces from the resulting torque (FGi)"""

        center_of_rotation = array(self.center_of_rotation)
        force_relative_to_G = array(self.force_location) - center_of_rotation
        torque_around_G = cross(force_relative_to_G, array(self.force))

        # zeroing torques not in the z direction (because they don't cause shear)
        torque_around_G[0] = 0
        torque_around_G[1] = 0

        rGi = [array(i) - center_of_rotation for i in self.fasteners_locations]

        force_times_area = [cross(torque_around_G, rGi[i]) * j for i, j in
                            enumerate(self.bolt_shear_area)]
        b = 0
        for i, j in enumerate(self.bolt_shear_area):
            b += j * norm(rGi[i]) ** 2

        return [a / b for a in force_times_area]

    @property
    def normal_stress(self):
        """normal stress in the bolt"""
        bolt_stress_area = [fastener.bolt.stress_area for fastener in self.fasteners]
        return self.bolt_load / bolt_stress_area

    @property
    def bolt_load(self):
        """tension force in the bolt (Fbj)"""
        return array(self.preloads) + (array(self.fastener_load) * array(self.fasteners_stiffness))

    @property
    def fastener_load(self):
        """the total normal force on bolt from the direct force and resulting bending moment (Pj)"""
        return [abs(i[2] + j[2]) for i, j in zip(self.direct_normal_load, self.bending_normal_load)]

    @property
    def direct_normal_load(self):
        """normal force directly from pure tension (PTj)"""
        normal_force = self.force[2]
        return [array([0, 0, (normal_force * stiffness) / sum(self.total_stiffness)]) for stiffness
                in self.total_stiffness]

    @property
    def neutral_point(self):
        """The location of the neutral point of tension"""
        bolts_x_locations = [bolt[0] for bolt in self.fasteners_locations]
        bolts_y_locations = [bolt[1] for bolt in self.fasteners_locations]
        bolts_z_locations = [bolt[2] for bolt in self.fasteners_locations]

        # location of the neutral point of tension
        Hx, Hy, Hz = 0, 0, 0
        for stiffness, location in zip(self.total_stiffness, bolts_x_locations):
            Hx += (stiffness * location) / sum(self.total_stiffness)
        for stiffness, location in zip(self.total_stiffness, bolts_y_locations):
            Hy += (stiffness * location) / sum(self.total_stiffness)
        for stiffness, location in zip(self.total_stiffness, bolts_z_locations):
            Hz += (stiffness * location) / sum(self.total_stiffness)
        return Hx, Hy, Hz

    @property
    def bending_normal_load(self):
        """normal force resulting from the bending moment (PMTVj)"""
        Hx, Hy, Hz = self.neutral_point

        # finding resulting moment relative to neutral center
        neutral_center = array([Hx, Hy, 0])
        force_relative_to_H = array(self.force_location) - neutral_center
        moment_around_H = cross(force_relative_to_H, array(self.force))
        # zeroing moment in the z direction (because it doesn't cause tension)
        moment_around_H[2] = 0

        # finding bolts distances from rotation edge
        edge_p1, edge_p2 = array(self.axis_of_rotation)
        edge_vector = edge_p2 - edge_p1
        perpendicular_edge_vector = [-edge_vector[1], edge_vector[0]]
        edge_direction = perpendicular_edge_vector / norm(perpendicular_edge_vector)
        distance_from_edge = [cross(edge_vector, p0[:-1] - edge_p1) / norm(edge_vector) *
                              edge_direction for p0 in array(self.fasteners_locations)]

        # edge = array(self.axis_of_rotation)
        # edge_direction = edge / norm(edge)
        # distance_from_edge = [edge + dot(r, edge_direction) * edge_direction for r in
        #                             array(self.fasteners_locations)]

        force_times_stiffness = [cross(moment_around_H, distance_from_edge[i]) * j for i, j in
                                 enumerate(self.total_stiffness)]

        b = 0
        for i, j in enumerate(self.total_stiffness):
            b += j * norm(distance_from_edge[i]) ** 2
        return [a / b for a in force_times_stiffness]

    @property
    def equivalent_stresses(self):
        """equivalent stresses according to Von-mises"""
        return [sqrt(normal_stress ** 2 + 3 * shear_stress ** 2) for normal_stress, shear_stress in
                zip(self.normal_stress, self.shear_stress)]

    def load_safety_factor(self, minimal_value=True, verbose=False):
        """Safety factor for loading (nL)
                :param bool minimal_value: If true returns the lowest safety factor of all the
                fasteners, if false returns a list of safety factors for each of the fasteners
                :param bool verbose: If true prints the safety value for each fastener
        """
        proof_loads = array([fastener.bolt.proof_load for fastener in self.fasteners])
        nL = (proof_loads - array(self.preloads)) / (self.bolt_load - array(self.preloads))
        if verbose:
            for i, fastener in enumerate(self.fasteners):
                print(f"{fastener} - nL = {nL[i]:.2f}")
            print("")
        return min(nL) if minimal_value else nL

    def separation_safety_factor(self, minimal_value=True, verbose=False):
        """Safety factor against fastener separation (n0)
                :param bool minimal_value: If true returns the lowest safety factor of all the
                fasteners, if false returns a list of safety factors for each of the fasteners
                :param bool verbose: If true prints the safety value for each fastener
        """
        n0 = array(self.preloads) / ((1 - array(self.fasteners_stiffness)) * self.fastener_load)
        if verbose:
            for i, fastener in enumerate(self.fasteners):
                print(f"{fastener} - n0 = {n0[i]:.2f}")
            print("")
        return min(n0) if minimal_value else n0

    def proof_safety_factor(self, minimal_value=True, verbose=False):
        """Safety factor for proof strength (np)
                :param bool minimal_value: If true returns the lowest safety factor of all the
                fasteners, if false returns a list of safety factors for each of the fasteners
                :param bool verbose: If true prints the safety value for each fastener
        """
        np = [fastener.bolt.proof_strength / eq for fastener, eq in
              zip(self.fasteners, self.equivalent_stresses)]
        if verbose:
            for i, fastener in enumerate(self.fasteners):
                print(f"{fastener} - np = {np[i]:.2f}")
            print("")
        return min(np) if minimal_value else np

    def variable_loading_stresses(self, Fmin, Fmax):
        """
        Returns the alternating and mean normal and shear stresses
            :param list[float] Fmin: Minimum force
            :param list[float] Fmax: Maximum force
            :return:A List of the alternating normal stress, a List of the alternating shear stress,
            a List of the mean normal stress and a List of the mean shear stress
            :rtype: dict{'alt_normal_stress': float,
                         'alt_shear_stress': float,
                         'mean_normal_stress': float,
                         'mean_shear_stress': float}
        """
        force = self.force  # saving the original force used in the class call

        # setting the pattern force attribute as the minimum force and
        # retrieving the minimum normal stress and the minimum shear stress
        # from the updated class properties
        self.force = Fmin
        min_normal_stress = array(self.normal_stress)
        min_shear_stress = array(self.shear_stress)

        # setting the pattern force attribute as the maximum force and
        # retrieving the maximum normal stress and the maximum shear stress
        # from the updated class properties
        self.force = Fmax
        max_normal_stress = array(self.normal_stress)
        max_shear_stress = array(self.shear_stress)

        alt_normal_stress = (max_normal_stress - min_normal_stress) / 2
        alt_shear_stress = (max_shear_stress - min_shear_stress) / 2
        mean_normal_stress = (max_normal_stress + min_normal_stress) / 2
        mean_shear_stress = (max_shear_stress + min_shear_stress) / 2

        self.force = force  # restoring the original force used in the class call

        return {'alt_normal_stress': alt_normal_stress,
                'alt_shear_stress': alt_shear_stress,
                'mean_normal_stress': mean_normal_stress,
                'mean_shear_stress': mean_shear_stress}

    def variable_equivalent_stresses(self, endurance_limit, Fmin, Fmax):
        """Returns the mean and alternating equivalent stress (σ_eq_a and σ_eq_m)
            :param list[EnduranceLimit] endurance_limit: List of the pattern's bolts
            endurance limit objects
            :param list[float] Fmin: Minimum force
            :param list[float] Fmax: Maximum force

            :returns: list of mean and alternating equivalent stresses pairs for each of the
            pattern's bolts
            :rtype: list[dict{'mean': float, 'alt': float}]
        """
        alt_normal_stress, alt_shear_stress, mean_normal_stress, mean_shear_stress = \
            self.variable_loading_stresses(Fmin, Fmax).values()
        variable_eq_stresses = []
        for i, fastener in enumerate(self.fasteners):
            analysis = FatigueAnalysis(endurance_limit=endurance_limit[i], ductile=True,
                                       Sy=fastener.bolt.yield_strength,
                                       Kf_normal=fastener.bolt.Kf,
                                       Kf_torsion=fastener.bolt.Kf,
                                       alt_normal_stress=alt_normal_stress[i],
                                       alt_torsion_stress=alt_shear_stress[i],
                                       mean_normal_stress=mean_normal_stress[i],
                                       mean_torsion_stress=mean_shear_stress[i])
            variable_eq_stresses.append({'mean': analysis.mean_eq_stress,
                                         'alt': analysis.alt_eq_stress})
        return variable_eq_stresses

    def fatigue_safety_factor(self, endurance_limits, Fmin, Fmax, verbose=False):
        """
        Returns the fatigue(Goodman) and static(Langar) safety factors for each bolt
        :param list[EnduranceLimit] endurance_limits: the EnduranceLimit object of each bolt
        :param list[float] Fmin: Minimum force
        :param list[float] Fmax: Maximum force
        :param bool verbose: Print additional information
        :return: A list of the fatigue and static safety factors pairs for each of the bolts
        :rtype: list[dict{'fatigue': float, 'static': float}]
        """
        variable_eq_stress = self.variable_equivalent_stresses(endurance_limits, Fmin, Fmax)

        safety_factors = []
        for i, fastener in enumerate(self.fasteners):
            preload_stress = fastener.preload / fastener.bolt.stress_area
            mean_stress = variable_eq_stress[i]['mean']
            alt_stress = variable_eq_stress[i]['alt']
            Sut = fastener.bolt.tensile_strength
            Se = endurance_limits[i].modified
            Sp = fastener.bolt.proof_strength
            if verbose:
                print(f"σi={preload_stress}, σa={alt_stress}, σm={mean_stress}, Sut={Sut}, Se={Se}")
            nf = ((Se * (Sut - preload_stress)) /
                  (Sut * alt_stress + Se * (mean_stress - preload_stress)))
            ns = Sp / (mean_stress + alt_stress)
            safety_factors.append({'fatigue': nf, 'static': ns})
        return safety_factors



"""A module containing the extension spring class"""
from math import pi

from me_toolbox.fatigue import FailureCriteria
from me_toolbox.springs import HelicalCompressionSpring
from me_toolbox.tools import percent_to_decimal


class ExtensionSpring(HelicalCompressionSpring):
    """An extension spring object"""

    def __repr__(self):
        return f"ExtensionSpring(max_force={self.max_force}, " \
               f"initial_tension={self.initial_tension}, wire_diameter={self.wire_diameter}, " \
               f"spring_diameter={self.diameter}, hook_r1={self.hook_r1}, " \
               f"hook_r2={self.hook_r2}, " \
               f"ultimate_tensile_strength={self.ultimate_tensile_strength}, " \
               f"shear_modulus={self.shear_modulus}, elastic_modulus={self.elastic_modulus}, " \
               f"body_shear_yield_percent={self.body_shear_yield_percent}, " \
               f"hook_normal_yield_percent={self.hook_normal_yield_percent}, " \
               f"hook_shear_yield_percent={self.hook_normal_yield_percent}, " \
               f"spring_rate={self.spring_rate}, shot_peened={self.shot_peened}, " \
               f"density={self.density}, working_frequency={self.working_frequency}, " \
               f"anchors={self.anchors})"

    def __str__(self):
        return f"ExtensionSpring(d={self.wire_diameter}, D={self.diameter}, " \
               f"k={self.spring_rate}, L={self.free_length})"

    def __init__(self, max_force, initial_tension, wire_diameter, spring_diameter, hook_r1, hook_r2,
                 ultimate_tensile_strength, shear_modulus, elastic_modulus,
                 body_shear_yield_percent, hook_normal_yield_percent, hook_shear_yield_percent,
                 spring_rate, shot_peened=False, density=None, working_frequency=None,
                 anchors=None):
        """Instantiate an extension spring object with the given parameters

        :param float max_force: The maximum load on the spring [N]
        :param float initial_tension: The initial tension in the spring [N]
        :param float wire_diameter: spring wire diameter [mm]
        :param float spring_diameter: spring diameter measured from
            the center point of the wire diameter [mm]
        :param float hook_r1: hook internal radius [mm]
        :param float hook_r2: hook bend radius [mm]
        :param float ultimate_tensile_strength: Ultimate tensile strength of the material [Mpa]
        :param float shear_modulus: Spring's material shear modulus [MPa]
        :param float elastic_modulus: Spring's material elastic modulus [MPa]
        :param float body_shear_yield_percent: Used to estimate the spring's body shear yield stress
        :param float hook_normal_yield_percent: Used to estimate the spring's hook yield stress
        :param float hook_shear_yield_percent: Used to estimate the spring's hook shear yield stress
        :param float spring_rate: K - spring constant [Nm/rad]
        :param bool shot_peened: if True adds to fatigue strength
        :param float or None density: Spring's material density [kg/m^3]
            (used for buckling and weight calculations)
        :param float or None working_frequency: the spring working frequency [Hz]
            (used for fatigue calculations)

        :returns: HelicalCompressionSpring
        """

        super().__init__(max_force, wire_diameter, spring_diameter, ultimate_tensile_strength,
                         body_shear_yield_percent, shear_modulus, elastic_modulus, end_type='plain',
                         spring_rate=spring_rate, set_removed=False, shot_peened=shot_peened,
                         density=density, working_frequency=working_frequency,
                         anchors=anchors, zeta=0.15)

        self.initial_tension = initial_tension
        self.hook_r1 = hook_r1
        self.hook_r2 = hook_r2
        self.body_shear_yield_percent = body_shear_yield_percent
        self.hook_normal_yield_percent = hook_normal_yield_percent
        self.hook_shear_yield_percent = hook_shear_yield_percent

        # self.check_design()

    def check_design(self, verbose=False):
        """Check if the spring index,active_coils,zeta and free_length
        are in acceptable range for good design

        :returns: True if all the checks are good
        :rtype: bool
        """
        good_design = True
        C = self.spring_index  # pylint: disable=invalid-name
        if isinstance(C, float) and not 3 <= C <= 16:
            print("Note: C - spring index should be in range of [3,16],"
                  "lower C causes surface cracks,\n"
                  "higher C causes the spring to tangle and requires separate packing")
            good_design = False

        active_coils = self.active_coils
        if isinstance(active_coils, float) and not 3 <= active_coils <= 15:
            print(f"Note: active_coils={active_coils:.2f} is not in range [3,15],"
                  f"this can cause non linear behavior")
            good_design = False

        if (self.density is not None) and (self.working_frequency is not None):
            natural_freq = self.natural_frequency
            if natural_freq <= 20 * self.working_frequency:
                print(
                    f"Note: the natural frequency={natural_freq} is less than 20*working"
                    f"frequency={20 * self.working_frequency}")
                good_design = False
        if verbose:
            print(f"good_design = {good_design}")

        return good_design

    @property
    def free_length(self):
        """Calculates the free length of the spring

        :returns: free_length - The free length
        :rtype: float
        """
        return 2 * (self.diameter - self.wire_diameter) + (self.body_coils + 1) * self.wire_diameter

    @property
    def solid_length(self):
        raise NotImplementedError("solid_length is inherited from HelicalCompressionSpring "
                                  "but has no use in ExtensionSpring")

    @property
    def Fsolid(self):
        raise NotImplementedError("Fsolid is inherited from HelicalCompressionSpring "
                                  "but has no use in ExtensionSpring")

    @property
    def body_coils(self):
        """Calculate active_coils which is the number of active coils (using Castigliano's theorem)

        :returns: number of active coils
        :rtype: float
        """
        return self.active_coils - (self.shear_modulus / self.elastic_modulus)

    @property
    def total_coils(self):
        raise NotImplementedError("total_coils is inherited from HelicalCompressionSpring "
                                  "but has no use in ExtensionSpring")

    @property
    def hook_normal_yield_strength(self):  # pylint: disable=invalid-name
        """getter for the yield strength attribute (Sy = % * Sut)

        :returns: hook bending yield strength
        :rtype: float
        """
        try:
            return (percent_to_decimal(self.hook_normal_yield_percent) *
                    self.ultimate_tensile_strength)
        except TypeError:
            return self.hook_normal_yield_percent * self.ultimate_tensile_strength

    @property
    def hook_shear_yield_strength(self):
        """getter for the yield strength attribute (Ssy = % * Sut)

        :returns: hook bending yield strength
        :rtype: float
        """
        try:
            return (percent_to_decimal(self.hook_shear_yield_percent) *
                    self.ultimate_tensile_strength)
        except TypeError:
            return self.hook_shear_yield_percent * self.ultimate_tensile_strength

    @property
    def hook_KA(self):  # pylint: disable=invalid-name
        """Returns The spring's bending stress correction factor

        :returns: Bending stress correction factor
        :rtype: float
        """
        C1 = 2 * self.hook_r1 / self.wire_diameter  # pylint: disable=invalid-name
        return ((4 * C1 ** 2) - C1 - 1) / (4 * C1 * (C1 - 1))

    @property
    def hook_KB(self):
        """Returns The spring's torsional stress correction factor

        :returns: Torsional stress correction factor
        :rtype: float
        """

        C2 = 2 * self.hook_r2 / self.wire_diameter  # pylint: disable=invalid-name
        return (4 * C2 - 1) / (4 * C2 - 4)

    @property
    def max_hook_normal_stress(self):
        """The normal stress due to bending and axial loads

        :returns: Normal stress
        :rtype: float
        """
        return self.calc_normal_stress(self.max_force)

    def calc_normal_stress(self, force):
        """Calculates the normal stress based on the max_force given.

        :param float force: Working max_force of the spring

        :returns: normal stress
        :rtype: float
        """
        return force * (self.hook_KA * (
                (16 * self.diameter) / (pi * self.wire_diameter ** 3)) + (
                                4 / (pi * self.wire_diameter ** 2)))

    @property
    def max_hook_shear_stress(self):
        """The spring's hook torsion stress

        :returns: Hook torsion stress
        :rtype: float
        """
        return self.calc_shear_stress(self.max_force, self.hook_KB)

    @property
    def max_body_shear_stress(self):
        """The spring's body torsion stress

        :returns: Body torsion stress
        :rtype: float
        """
        # return self.calc_max_shear_stress(self.max_force, hook=False)
        return self.calc_shear_stress(self.max_force, self.factor_Kw)

    @property
    def natural_frequency(self):
        raise NotImplementedError("Need to adapt the formula to extension spring")

    def calc_deflection(self, force):
        """Calculate the spring max_deflection (change in length) due to specific max_force.

        :param float force: Spring working max_force

        :returns: Spring max_deflection
        :rtype: float
        """
        return (force - self.initial_tension) / self.spring_rate

    def static_safety_factor(self):
        """ Returns the static safety factors for the hook (torsion and
        bending), and for the spring's body (torsion)

        :returns: Spring's body (torsion) safety factor, Spring's hook bending safety factor,
            Spring's hook torsion safety factor
        :type: dict{str: float}
        """

        n_body = self.shear_yield_strength / self.max_body_shear_stress
        n_hook_normal = self.hook_normal_yield_strength / self.max_hook_normal_stress
        n_hook_shear = self.hook_shear_yield_strength / self.max_hook_shear_stress

        return {'n_body': n_body, 'n_hook_normal': n_hook_normal, 'n_hook_shear': n_hook_shear}

    def fatigue_analysis(self, max_force, min_force, reliability,
                         criterion='gerber', verbose=False, metric=True):
        """Fatigue analysis of the hook section, for normal and shear stress,and for the
        body section for shear and static yield.

        :param float max_force: Maximal max_force acting on the spring
        :param float min_force: Minimal max_force acting on the spring
        :param float reliability: in percentage
        :param str criterion: fatigue criterion
        :param bool verbose: print more details
        :param bool metric: Metric or imperial

        :returns: Normal and shear safety factors for the hook section and
            static and dynamic safety factors for body section
        :rtype: dict[str, float]
        """
        # calculating mean and alternating forces
        alt_force = abs(max_force - min_force) / 2
        mean_force = (max_force + min_force) / 2

        # calculating mean and alternating stresses for the hook section
        # shear stresses:
        alt_shear_stress = self.calc_shear_stress(alt_force, self.hook_KB)
        mean_shear_stress = (mean_force / alt_force) * alt_shear_stress
        # normal stresses due to bending:
        alt_normal_stress = self.calc_normal_stress(alt_force)
        mean_normal_stress = (mean_force / alt_force) * alt_normal_stress

        Sse = self.shear_endurance_limit(reliability, metric)  # pylint: disable=invalid-name
        Ssu = self.shear_ultimate_strength
        Ssy_body = self.shear_yield_strength
        Ssy_hook = self.hook_shear_yield_strength
        Sy_hook = self.hook_normal_yield_strength
        Se = Sse / 0.577  # estimation using distortion-energy theory
        Sut = self.ultimate_tensile_strength

        try:
            nf_hook_normal, _ = FailureCriteria.get_safety_factors(Sy_hook, Sut, Se,
                                                                   alt_normal_stress,
                                                                   mean_normal_stress, criterion)

            nf_hook_shear, _ = FailureCriteria.get_safety_factors(Ssy_hook, Ssu, Sse,
                                                                  alt_shear_stress,
                                                                  mean_shear_stress, criterion)
        except TypeError as typ_err:
            raise ValueError(f"Fatigue analysis can't handle symbolic vars") from typ_err

        # calculating mean and alternating stresses for the body section
        # shear stresses:
        alt_body_shear_stress = self.calc_shear_stress(alt_force, self.hook_KB)
        mean_body_shear_stress = (mean_force / alt_force) * alt_shear_stress

        nf_body, ns_body = FailureCriteria.get_safety_factors(Ssy_body, Ssu, Sse,
                                                              alt_body_shear_stress,
                                                              mean_body_shear_stress, criterion)

        if verbose:
            print(f"Alternating force = {alt_force}, Mean force = {mean_force}\n"
                  f"Alternating shear stress = {alt_shear_stress},"
                  f"Mean shear stress = {mean_shear_stress}\n"
                  f"Alternating normal stress = {alt_normal_stress},"
                  f"Mean normal stress = {mean_normal_stress}\n"
                  f"Alternating body shear stress = {alt_body_shear_stress},"
                  f"Mean body shear stress = {mean_body_shear_stress}\n"
                  f"Sse = {Sse}, Se = {Se}")

        return {'nf_body': nf_body, 'ns_body': ns_body, 'nf_hook_normal': nf_hook_normal,
                'nf_hook_shear': nf_hook_shear}

    def min_wire_diameter(self, Ap, m, safety_factor, spring_index=None):
        """The minimal wire diameters (for shear and normal stresses)
        for given safety factor in order to avoid failure,

        Because KA and KB contains d no simple solution is available as in the
        HelicalCompressionSpring, so we assume an initial K and iterate until convergence,
        be aware that for some static_safety_factor convergence my not occur.

        NOTE: for static use only.

        :param float safety_factor: Static safety factor
        :param float spring_index: Spring index

        :returns: The minimal wire diameter
        :rtype: tuple(float, float)
        """
        F = self.max_force
        C = spring_index

        factor_k, temp_k = 1.1, 0
        normal_diam = 0
        while abs(factor_k - temp_k) > 1e-4:
            # waiting for k to converge
            percent = self.hook_normal_yield_percent
            normal_diam = (safety_factor * F * (16 * factor_k * C - 4) / (percent * Ap * pi)) ** (
                    1 / (2 - m))
            temp_k = factor_k
            factor_k = ((16 * self.hook_r1 ** 2 - 2 * self.hook_r1 * normal_diam - normal_diam ** 2)
                        / (16 * self.hook_r1 ** 2 - 8 * self.hook_r1 * normal_diam))

        factor_k, temp_k = 1.1, 0
        shear_diam = 0
        while abs(factor_k - temp_k) > 1e-4:
            # waiting for k to converge
            percent = self.hook_shear_yield_percent
            shear_diam = ((8 * factor_k * F * C * safety_factor) / (percent * Ap * pi)) ** (
                    1 / (2 - m))
            temp_k = factor_k
            factor_k = (8 * self.hook_r2 - shear_diam) / (8 * self.hook_r2 - 4 * shear_diam)

        try:
            return max(normal_diam, shear_diam)
        except TypeError:
            return normal_diam, shear_diam

    def min_spring_diameter(self, static_safety_factor):
        """return the minimum spring diameter to avoid static failure
        according to the given safety factor.

        :param float static_safety_factor: factor of safety

        :returns: The minimal spring diameter
        :rtype: float
        """
        # extracted from shear stress
        diameter_shear = (self.hook_shear_yield_strength * pi * self.wire_diameter ** 3) / (
                self.hook_KB * 8 * self.max_force * static_safety_factor)
        # extracted from normal stress
        diameter_normal = (1 / (4 * self.hook_KA)) * \
                          (((self.hook_normal_yield_strength * pi * self.wire_diameter ** 3) /
                            (4 * self.max_force * static_safety_factor)) - self.wire_diameter)
        try:
            return max(diameter_shear, diameter_normal)
        except TypeError:
            return diameter_shear, diameter_normal

    def buckling(self, anchors, verbose=True):
        raise NotImplementedError("Inherited from HelicalCompressionSpring but useless here")

    @staticmethod
    def calc_spring_rate(wire_diameter, spring_diameter, shear_modulus, active_coils):
        raise NotImplementedError("Didn't get to it yet")



"""module containing the FatigueAnalysis class and
calc_kf for calculating dynamic stress concentration factor
"""
from math import log10
from sympy import sqrt
from me_toolbox.tools import NotInRangeError, print_atributes
from sympy import oo
from me_toolbox.fatigue import FailureCriteria


class FatigueAnalysis:
    """Perform fatigue analysis"""

    # TODO: make endurance_limit optional and add option to input Se, Stress_type and Kc directly
    # TODO: replace Kc with some other way to determine what to do in factors calculation

    def __init__(self, endurance_limit, ductile, Sy=None,
                 Kf_bending=0, Kf_normal=0, Kf_torsion=0,
                 alt_bending_stress=0, alt_normal_stress=0, alt_torsion_stress=0,
                 mean_bending_stress=0, mean_normal_stress=0, mean_torsion_stress=0):
        """ Instantiating fatigue object
        Note: all stresses are in [MPa]
        :param EnduranceLimit endurance_limit: EnduranceLimit object containing
        the modified Se
        :param bool ductile: True if material is ductile
        :param float Sy: Yield strength in [Mpa]
        :param float Kf_bending: dynamic stress concentration factor for bending
        :param Kf_normal: dynamic stress concentration factor for normal
        :param Kf_torsion: dynamic stress concentration factor for torsion
        :param float alt_bending_stress: Alternating bending stress
        :param float alt_normal_stress: Alternating normal stress
        :param float alt_torsion_stress: Alternating torsion stress
        :param float mean_bending_stress: Mean bending stresses
        :param float mean_normal_stress: Mean normal stresses
        :param float mean_torsion_stress: Mean torsion stresses
        """

        self.Sy = Sy
        self.Se = endurance_limit.modified
        self.Kc = endurance_limit.Kc
        self.Sut = endurance_limit.Sut
        self.stress_type = endurance_limit.stress_type
        self.ductile = ductile
        self.Kf_bending, self.Kf_normal, self.Kf_torsion = Kf_bending, Kf_normal, Kf_torsion
        self.alt_bending_stress = alt_bending_stress
        self.alt_normal_stress = alt_normal_stress
        self.alt_torsion_stress = alt_torsion_stress
        self.mean_bending_stress = mean_bending_stress
        self.mean_normal_stress = mean_normal_stress
        self.mean_torsion_stress = mean_torsion_stress
        self.alt_eq_stress = self.calc_alt_eq_stress()
        self.mean_eq_stress = self.calc_mean_eq_stress()

    @staticmethod
    def calc_kf(q, Kt):
        """Calculating the dynamic stress concentration factor (Kf)
        (this calculation is the same for the shear concentration factor)

        :param float Kt: stress concentration theoretical factor
        :param float q: notch Sensitivity

        :returns: dynamic stress concentration factor
        :rtype: float
        """
        return 1 + q * (Kt - 1)

    @staticmethod
    def calc_thread_kf(grade, manufacturing):
        """Calculating the dynamic stress concentration factor (Kf) for threads

        :param float grade: The grade of the bolt
        :param string manufacturing: Manufacturing methode 'Rolled Threads' or 'Cut Threads'

        :returns: dynamic stress concentration factor for threads
        :rtype: float
        """
        if manufacturing != "Rolled Threads" and manufacturing != "Cut Threads":
            raise ValueError(f"The only exceptable manufacturing methods are 'Rolled Threads' and "
                             f"'Cut Threads'")
        if 3.6 <= grade <= 5.8:
            options = {"Rolled Threads": 2.2, "Cut Threads": 2.8}
        elif 6.6 <= grade <= 10.9:
            options = {"Rolled Threads": 3, "Cut Threads": 3.8}
        else:
            raise ValueError(f"Grade entered is {grade} but grade can only be in the range of"
                             f"3.6 to 5.8 or 6.6 to 0.9")
        return options[manufacturing]

    def calc_alt_eq_stress(self):
        """Returns the alternating equivalent stress according to the load type indicated by Kc

        :returns: Alternating equivalent stress
        :rtype: float
        """

        Kc = self.Kc
        if Kc == 1 and self.stress_type == 'multiple':
            corrected_bending = self.Kf_bending * self.alt_bending_stress
            corrected_normal = self.Kf_normal * (self.alt_normal_stress / 0.85)
            corrected_torsion = self.Kf_torsion * self.alt_torsion_stress
            return sqrt((corrected_bending + corrected_normal) ** 2 + 3 * corrected_torsion ** 2)

        elif Kc == 1:
            return self.Kf_bending * self.alt_bending_stress

        elif Kc == 0.85:
            return self.Kf_normal * self.alt_normal_stress

        elif Kc == 0.59:
            return self.Kf_torsion * self.alt_torsion_stress

    def calc_mean_eq_stress(self):
        """Returns the mean equivalent stress according to the load type indicated by Kc

        :returns: Mean equivalent stress
        :rtype: float
        """
        if self.ductile:
            # if the material is ductile no correction is needed
            Kf_bending, Kf_normal, Kf_torsion = 1, 1, 1
        else:
            Kf_bending, Kf_normal, Kf_torsion = self.Kf_bending, self.Kf_normal, self.Kf_torsion

        Kc = self.Kc
        if Kc == 1 and self.stress_type == 'multiple':
            corrected_bending = Kf_bending * self.mean_bending_stress
            corrected_normal = Kf_normal * self.mean_normal_stress
            corrected_torsion = Kf_torsion * self.mean_torsion_stress

            return sqrt((corrected_bending + corrected_normal) ** 2 + 3 * corrected_torsion ** 2)

        elif Kc == 1:
            return Kf_bending * self.mean_bending_stress

        elif Kc == 0.85:
            return Kf_normal * self.mean_normal_stress

        elif Kc == 0.59:
            return Kf_torsion * self.mean_torsion_stress

    @property
    def shear_ultimate_strength(self):
        """Returns shear_ultimate_strength which is the
        ultimate_tensile_strength correction for shear stress

        :returns: Sst - ultimate shear tensile strength
        :type: float
        """
        return 0.67 * self.Sut

    @property
    def shear_yield_stress(self):
        """Returns shear_yield_strength which is the Sy correction for shear stress

        :returns: shear_yield_strength - yield stress for shear
        :type: float
        """
        return self.Sy / sqrt(3).evalf()

    @property
    def modified_goodman(self):
        """Safety factor according to modified Goodman failure criterion
        (very common criterion)

        :returns: Safety factor
        :rtype: any

        :raises ValueError: if ultimate_tensile_strength is not in the fatigue call
        """

        if self.mean_eq_stress < 0:
            return None

        ultimate_strength = self.Sut
        if self.Kc == 0.59:
            # ultimate_tensile_strength correction for shear stress
            ultimate_strength = self.shear_ultimate_strength

        return FailureCriteria.modified_goodman(ultimate_strength, self.Se,
                                                self.alt_eq_stress,
                                                self.mean_eq_stress)

    @property
    def soderberg(self):
        """Safety factor according to Soderberg failure criterion
        (the safest criterion)

        :returns: Safety factor
        :rtype: any

        :raises ValueError: if Sy is not in the fatigue call
        """

        if self.mean_eq_stress < 0:
            return None

        yield_strength = self.Sy
        if self.Kc == 0.59:
            # ultimate_tensile_strength correction for shear stress
            yield_strength = self.shear_yield_stress

        return FailureCriteria.soderberg(yield_strength, self.Se,
                                         self.alt_eq_stress,
                                         self.mean_eq_stress)

    @property
    def gerber(self):
        """Safety factor according to Gerber failure criterion
        (the most lenient criterion)

        :returns: Safety factor
        :rtype: any
        :raises ValueError: if ultimate_tensile_strength is not in the fatigue call
        """

        if self.mean_eq_stress < 0:
            return None

        ultimate_strength = self.Sut
        if self.Kc == 0.59:
            # ultimate_tensile_strength correction for shear stress
            ultimate_strength = self.shear_ultimate_strength

        return FailureCriteria.gerber(ultimate_strength, self.Se,
                                      self.alt_eq_stress,
                                      self.mean_eq_stress)

    @property
    def ASME_elliptic(self):
        """Safety factor according to ASME Failure criterion

        :returns: Safety factor
        :rtype: any

        :raises ValueError: if ultimate_tensile_strength is not in the fatigue call
        """

        if self.mean_eq_stress < 0:
            return None

        yield_strength = self.Sut
        if self.Kc == 0.59:
            # ultimate_tensile_strength correction for shear stress
            yield_strength = self.shear_yield_stress

        return FailureCriteria.asme_elliptic(yield_strength, self.Se, self.alt_eq_stress,
                                             self.mean_eq_stress)

    @property
    def langer_static_yield(self):
        """ Static safety factor according to Langer Failure criterion
        it's customary to use Langer, as an assessment to yielding in the first cycle

        :returns: Safety factor
        :rtype: any

        :raises ValueError: if Sy is not in the fatigue call
        """

        yield_strength = self.Sy
        if self.Kc == 0.59:
            # yield_strength correction for shear stress
            Ssy = 0.67 * self.Sut
            yield_strength = Ssy

        return FailureCriteria.langer_static_yield(yield_strength,
                                                   self.alt_eq_stress,
                                                   self.mean_eq_stress)

    def get_safety_factors(self, criterion, verbose=False):
        """Returns dynamic and static safety factors
        according to the quadrant in the alternating-mean
        stress plain where the stresses are in

        Note: Should always be used instead of accessing the
        individual safety factors properties directly because it takes into account the
        stress direction

        :param str criterion: The criterion to use (modified goodman, soderberg, gerber,
         asme-elliptic)
        :param bool verbose: Print the result

        :returns: dynamic and static safety factors
        :rtype: tuple[float, float]
        """

        return FailureCriteria.get_safety_factors(self.Sy, self.Sut, self.Se, self.alt_eq_stress,
                                                  self.mean_eq_stress, criterion, verbose)

    @property
    def Sm_stress(self):
        """Getter for the Sm_stress (at 1e3 cycles) property

        :returns: Sm - stress at 1e3 cycles
        :rtype: float
        """
        return self.calc_Sm(self.Sut)

    @staticmethod
    def calc_Sm(Sut):
        """Calculate Sm_stress which is the stress at 1e3 cycles, the boundary
        dividing Low cycle fatigue and high cycle fatigue

        :param Sut: Ultimate tensile strength

        :returns: Sm_stress stress
        :rtype: float
        """

        def f(x):
            """ f - fatigue strength fraction
                function constructed from curve fitting to the f graph in Shigley's
                the range of the fit is ( 70[kPsi] < ultimate_tensile_strength < 200[kPsi] ) """
            return (-2.56710686e-16 * x ** 5 + 1.35729780e-12 * x ** 4 - 2.92474777e-09 * x ** 3 +
                    3.28990748e-06 * x ** 2 - 2.04929617e-03 * x + 1.38405394e+00)

        # TODO: add proper warning and solutions to ultimate_tensile_strength out of graph range
        if Sut < 482.633:  # 482.633[Mpa] = 70[kPsi]
            print(f"Note: ultimate_tensile_strength={Sut} < 482.633[Mpa] (70[kPsi]) so f~0.9")
            return 0.9 * Sut
        elif Sut > 1378.95:
            print(
                f"Note: ultimate_tensile_strength={Sut} > 1378.95[Mpa] (200[kPsi]) which is out of "
                f"the graph range, f={f(Sut)}")
        return f(Sut) * Sut

    def num_of_cycle(self, z=-3):
        """ calculate number of cycles until failure

        Note: zeta = log(N1) - log(N2), N1 - number of cycles at Sm_stress,
            N2 - Number of cycles at Se for steel N1=1e3 and N2 = 1e6

        :param float z: -3 for steel where N=1e6, -5 for metal where N=1e8,
            -5.69 for metal where N=5e8

        :returns: The Number of cycles and the fatigue stress at failure
        :rtype: tuple[float, float]
        """
        mean_stress = self.mean_eq_stress
        alternating_stress = self.alt_eq_stress
        Se = self.Se
        Sut, Sy, Sm = self.Sut, self.Sy, self.Sm_stress

        if mean_stress >= 0:
            reversible_stress = alternating_stress / (1 - (mean_stress / Sut))
        else:
            reversible_stress = alternating_stress

        if Sm < reversible_stress < Sy:
            # Low Cycle Fatigue
            if z != -3:
                raise ValueError(
                    "Number of cycles calculation for low cycle fatigue is only possible for "
                    "zeta=-3 ")

            a = Sut
            b = (1 / z) * log10(Sut / Sm)

        elif Se < reversible_stress < Sm:
            # High Cycle Fatigue
            a = Sm * (Sm / Se) ** (-3 / z)
            b = (1 / z) * log10(Sm / Se)

        elif reversible_stress < Se:
            print(
                f"reversible_stress = {reversible_stress} < Se = {Se}, Number of cycles is "
                f"infinite")
            return oo  # TODO: add to documentation the meaning of oo
        else:
            raise NotInRangeError("Reversible Stress", reversible_stress,
                                  (f'Sy={Sy}', f'Sm_stress={Sm}', f'Se={Se}'))
        N = (reversible_stress / a) ** (1 / b)
        return N, a * N ** b

    def miner_rule(self, stress_groups, Sut, Se, Sy=None, z=-3, verbose=False,
                   alt_mean=False, freq=False):
        """ Calculates total number of cycles for multiple periodic loads,
        the stress_groups format is as follows:
        [number_of_repetitions, maximum_stress, minimum_stress]

        Note: number_of_repetitions = frequency [Hz] * time

        Note: if the material don't have fatigue limit use the fatigue strength at Se=Sf(N=1e8)

        :param list stress_groups: list containing the pick stresses and number of repetition
        :param float Sut: Ultimate tensile strength [MPa]
        :param float Sy: yield strength [MPa], if None only HCF is checked
        :param float Se: endurance limit [MPa]
        :param float z: -3 for steel where N=1e6, -5 for metal where N=1e8, -5.69 for a metal
            where N=5e8
        :param bool verbose: printing the groups
            [number_of_repetitions,maximum_stress, minimum_stress, reversible_stress, Number of
            cycles]
        :param bool freq: if the input is frequency instead of number of repetition
        :param bool alt_mean: if True the stress_group structure contains
            alternating and mean stresses: [number_of_repetitions, alternating_stress, mean_stress]
            instead of the max and min stresses:
            [number_of_repetitions, maximum_stress, minimum_stress]

        :returns: Total number of cycles
        :rtype: float
        """

        Sm = self.calc_Sm(Sut)
        for group in stress_groups:

            # if the stress given are minimum and maximum instead of alternating and mean
            if not alt_mean:
                mean_stress = (group[1] + group[2]) / 2
                alternating_stress = abs(group[1] - group[2]) / 2
            else:
                alternating_stress = group[1]
                mean_stress = group[2]

            # calculate the reversible stress according to the mean stress sign
            if mean_stress >= 0:
                reversible_stress = alternating_stress / (1 - (mean_stress / Sut))
            else:
                reversible_stress = alternating_stress

            group.append(reversible_stress)

            if ((Sy is not None) and (Sm < reversible_stress < Sy)) or reversible_stress < Se:
                # infinite num of cycle - either the stress is less then
                # the endurance limit or its low cycle fatigue
                group.append(oo)

            elif Se < reversible_stress < Sm:
                # High Cycle Fatigue
                # TODO: this calculation is the same every
                #   iteration consider change to internal function and implement cash
                a = Sm * (Sm / Se) ** (-3 / z)
                b = (1 / z) * log10(Sm / Se)
                N = (reversible_stress / a) ** (1 / b)
                group.append(N)

            else:
                # print error but don't stop the loop
                print(
                    f"Reversible Stress = {reversible_stress} not in range,"
                    f"LCF-range=(Sm_stress={Sm},Sy={Sy}), "
                    f"HCF-range(Se={Se},Sm_stress={Sm})")

        result = 0
        for group in stress_groups:
            # summing n/N
            result += (group[0] / group[-1])
            if verbose:
                print(group)

        # casting the result to float - my be needed because the use of sympy's oo
        N_total = float(1 / result)

        if verbose:
            if freq:
                print(f"total time = {N_total:.2f} [s]")
            else:
                print(f"N_total = {N_total:.2f}")
        return N_total

    def get_info(self):
        """print object attributes"""
        print_atributes(self)
